#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import re
import base64
import collections
import hashlib
import sys
import datetime
import pyasn1.codec.der.decoder
from pyasn1_modules import rfc2459, rfc2437
import x509sak.ASN1Models
from x509sak.BaseAction import BaseAction
from x509sak.ScrapeEngine import ScrapeEngine
from x509sak.Tools import PEMDataTools, JSONTools
from x509sak.KwargsChecker import KwargsChecker
from x509sak.Intervals import Intervals, Interval

class _DERSanityCheck():
	class SanityCheckFailedException(Exception): pass

	@classmethod
	def check_rsa_key(cls, asn1):
		if asn1["prime1"] * asn1["prime2"] != asn1["modulus"]:
			raise cls.SanityCheckFailedException("Product p * q is not equal to modulus n; likely not a valid RSA private key.")

	@classmethod
	def check_ec_key(cls, asn1):
		if len(asn1["privateKey"]) <= 8:
			raise cls.SanityCheckFailedException("Short (%d bytes) private key scalar; likely not a valid ECC private key." % (len(asn1["privateKey"])))

	@classmethod
	def check_dsa_sig(cls, asn1):
		r_bitlen = int(asn1["r"]).bit_length()
		if r_bitlen <= 64:
			raise cls.SanityCheckFailedException("Short (%d bit) DSA r value; likely not a valid DSA signature." % (r_bitlen))

class ActionScrapeStats():
	_FRIENDLY_DATA_TYPE = {
		"crt":				"X.509 Certificate",
		"openssh_key":		"OpenSSH Private Key",
		"dsa_key":			"DSA Private Key",
		"rsa_key":			"RSA Private Key",
		"ec_key":			"ECC Private Key",
		"pubkey":			"Public Key",
		"crl":				"Certificate Revocation List",
		"csr":				"Certificate Signing Request",
		"unknown":			"Unknown",
	}
	_Finding = collections.namedtuple("Finding", [ "offset", "length", "data_type", "extension", "action", "filename" ])

	def __init__(self, args):
		self._args = args
		self._findings = [ ]
		self._start_time = datetime.datetime.utcnow()
		self._end_time = None
		self._end_offset = None
		self._active_der_types = None
		self._pem_potential_match = 0
		self._pem_successful_decode = 0
		self._der_potential_match = 0
		self._der_attempted_decode = 0
		self._der_successful_decode = 0
		self._der_passed_plausibility = 0
		self._der_failed_plausibility = 0

	def set_active_der_types(self, active_der_types):
		self._active_der_types = active_der_types

	def pem_potential_match(self):
		self._pem_potential_match += 1

	def pem_successful_decode(self):
		self._pem_successful_decode += 1

	def der_potential_match(self):
		self._der_potential_match += 1

	def der_attempt_decode(self):
		self._der_attempted_decode += 1

	def der_successful_decode(self):
		self._der_successful_decode += 1

	def der_passed_plausibility(self):
		self._der_passed_plausibility += 1

	def der_failed_plausibility(self):
		self._der_failed_plausibility += 1

	def record_finding(self, offset, length, data_type, extension, action, filename = None):
		self._findings.append(self._Finding(offset = offset, length = length, data_type = data_type, extension = extension, action = action, filename = filename))

	def finish(self, end_offset):
		self._end_time = datetime.datetime.utcnow()
		self._end_offset = end_offset

	def as_dict(self):
		return {
			"meta": {
				"filename":			self._args.filename,
				"start_offset":		self._args.seek_offset,
				"end_offset":		self._end_offset,
				"start_time_utc":	self._start_time,
				"end_time_utc":		self._end_time,
				"time_secs":		round((self._end_time - self._start_time).total_seconds()),
			},
			"analysis": {
				"pem": {
					"potential_match":		self._pem_potential_match,
					"successful_decode":	self._pem_successful_decode,
				},
				"der": {
					"active_types":			self._active_der_types,
					"potential_match":		self._der_potential_match,
					"attempted_decode":		self._der_attempted_decode,
					"successful_decode":	self._der_successful_decode,
					"passed_plausibility":	self._der_passed_plausibility,
					"failed_plausibility":	self._der_failed_plausibility,
				},
			},
			"findings": [ {
				"offset":		finding.offset,
				"length":		finding.length,
				"data_type":	finding.data_type,
				"extension":	finding.extension,
				"action":		finding.action,
			} for finding in self._findings	],
		}

	def dump(self, f = None):
		if f is None:
			f = sys.stderr
		stats = self.as_dict()
		print("Statistics of scraping %s" % (stats["meta"]["filename"]), file = f)
		print(file = f)
		print("Range      : From 0x%x to 0x%x" % (stats["meta"]["start_offset"], stats["meta"]["end_offset"]), file = f)
		print("Time       : From %s UTC to %s UTC" % (stats["meta"]["start_time_utc"].strftime("%Y-%m-%d %H:%M:%S"), stats["meta"]["end_time_utc"].strftime("%Y-%m-%d %H:%M:%S")), file = f)
		data_mib = (stats["meta"]["end_offset"] - stats["meta"]["start_offset"]) / 1024 / 1024

		if stats["meta"]["time_secs"] > 0.1:
			avg = "%.1f MiB/sec" % (data_mib / stats["meta"]["time_secs"])
		else:
			avg = "N/A"
		print("Processed  : %.0f MiB, average speed %s" % (data_mib, avg), file = f)
		print("PEM matches: %d offsets analyzed, %d successful PEM decodings" % (stats["analysis"]["pem"]["potential_match"], stats["analysis"]["pem"]["successful_decode"]), file = f)
		print("DER matches: %d offsets analyzed, %d attempted DER decodings total, %d successful DER decodings, %d also passed plausibility check" % (stats["analysis"]["der"]["potential_match"], stats["analysis"]["der"]["attempted_decode"], stats["analysis"]["der"]["successful_decode"], stats["analysis"]["der"]["passed_plausibility"]), file = f)
		print("DER types  : %s" % (", ".join(stats["analysis"]["der"]["active_types"])), file = f)
		written = sum(1 for finding in stats["findings"] if finding["action"] == "written")
		zero_length = sum(1 for finding in stats["findings"] if finding["action"] == "discard:zero_length")
		non_unique = sum(1 for finding in stats["findings"] if finding["action"] == "discard:non_unique")
		nested = sum(1 for finding in stats["findings"] if finding["action"] == "discard:nested")
		print("Findings   : %d total, %d written to disk, %d discarded" % (len(stats["findings"]), written, len(stats["findings"]) - written), file = f)
		print("Discards   : %d discarded because zero-length, %d because nested, %d because non-unique" % (zero_length, nested, non_unique), file = f)

		if written > 0:
			print(file = f)
			print("%-30s %s" % ("Data type", "Found"), file = f)
			stats_by_type = collections.Counter(finding["data_type"] for finding in stats["findings"] if finding["action"] == "written")
			for (data_type, counter) in stats_by_type.items():
				print("%-30s %d" % (self._FRIENDLY_DATA_TYPE.get(data_type, data_type), counter), file = f)

class ActionScrape(BaseAction):
	_DERHandler = collections.namedtuple("DERHandler", [ "asn1_spec", "data_type", "extension", "pem_marker", "sanity_check_fn", "precedence" ])
	_PEM_BEGIN = re.compile("^-----BEGIN (?P<marker>[ A-Za-z0-9]+)-----")
	_MARKERS = {
		"CERTIFICATE":				"crt",
		"OPENSSH PRIVATE KEY":		"openssh_key",
		"DSA PRIVATE KEY":			"dsa_key",
		"RSA PRIVATE KEY":			"rsa_key",
		"EC PRIVATE KEY":			"ec_key",
		"PUBLIC KEY":				"pubkey",
		"X509 CRL":					"crl",
		"CERTIFICATE REQUEST":		"csr",
		"NEW CERTIFICATE REQUEST":	"csr",
	}
	_DER_CLASSES = {
		handler_class.data_type: handler_class for handler_class in (
			_DERHandler(asn1_spec = rfc2459.Certificate(), data_type = "crt", extension = "der", pem_marker = "CERTIFICATE", sanity_check_fn = None, precedence = 10),
			_DERHandler(asn1_spec = rfc2437.RSAPrivateKey(), data_type = "rsa_key", extension = "der", pem_marker = "RSA PRIVATE KEY", sanity_check_fn = _DERSanityCheck.check_rsa_key, precedence = 20),
			_DERHandler(asn1_spec = rfc2459.DSAPrivateKey(), data_type = "dsa_key", extension = "der", pem_marker = "DSA PRIVATE KEY", sanity_check_fn = None, precedence = 20),
			_DERHandler(asn1_spec = rfc2459.SubjectPublicKeyInfo(), data_type = "pubkey", extension = "der", pem_marker = "PUBLIC KEY", sanity_check_fn = None, precedence = 30),
			_DERHandler(asn1_spec = x509sak.ASN1Models.ECPrivateKey(), data_type = "ec_key", extension = "der", pem_marker = "EC PRIVATE KEY", sanity_check_fn = _DERSanityCheck.check_ec_key, precedence = 20),
			_DERHandler(asn1_spec = x509sak.ASN1Models.PFX(), data_type = "pkcs12", extension = "p12", pem_marker = None, sanity_check_fn = None, precedence = 0),
			_DERHandler(asn1_spec = x509sak.ASN1Models.DSASignature(), data_type = "dsa_sig", extension = "der", pem_marker = None, sanity_check_fn = _DERSanityCheck.check_dsa_sig, precedence = 40),
	)}
	handler_classes = sorted(list(_DER_CLASSES.keys()))

	def __init__(self, cmdname, args):
		BaseAction.__init__(self, cmdname, args)

		# Plausibilize input parameters
		kwargs_checker = KwargsChecker(optional_arguments = set(self._DER_CLASSES.keys()))
		kwargs_checker.check(self._args.include_dertype, hint = "DER classes to be included")
		kwargs_checker.check(self._args.exclude_dertype, hint = "DER classes to be excluded")

		# Plausibilize output directory
		if os.path.exists(self._args.outdir) and (not self._args.force):
			raise Exception("Directory %s already exists. Remove it first or use --force." % (self._args.outdir))
		try:
			os.makedirs(self._args.outdir)
		except FileExistsError:
			pass

		# Determine active DERHandler classes
		if len(self._args.include_dertype) == 0:
			active_der_types = set(self._DER_CLASSES.keys())
		else:
			active_der_types = set(self._args.include_dertype)
		active_der_types -= set(self._args.exclude_dertype)
		self._active_der_types = [ self._DER_CLASSES[class_name] for class_name in active_der_types ]
		self._active_der_types.sort(key = lambda handler: (handler.precedence, handler.data_type))

		self._stats = ActionScrapeStats(self._args)
		self._stats.set_active_der_types([ handler_class.data_type for handler_class in self._active_der_types ])

		self._matches = Intervals()
		self._hashes = set()
		engine = ScrapeEngine(self._args.filename)
		if not self._args.no_pem:
			engine.search(self._find_pem, b"-----BEGIN ", min_length = 52, max_length = 32 * 1024)
		if (not self._args.no_der) and (len(self._active_der_types) > 0):
			self._log.debug("Looking for %d DER type(s): %s", len(self._active_der_types), ", ".join(handler.data_type for handler in self._active_der_types))
			engine.search(self._find_der, bytes.fromhex("30"), min_length = 2, max_length = 32 * 1024)
		end_offset = engine.commence(start_offset = self._args.seek_offset, length = self._args.analysis_length, progress_callback = self._progress_callback)
		self._stats.finish(end_offset)
		self._stats.dump()
		if self._args.write_json is not None:
			JSONTools.write_to_file(self._stats.as_dict(), self._args.write_json)

	def _progress_callback(self, position, total_length, elapsed_secs):
		self._log.debug("Scan at %.0f MiB of %.0f MiB, %.1f%%. Average speed %.1f MiB/sec", position / 1024 / 1024, total_length / 1024 / 1024, position / total_length * 100, position / 1024 / 1024 / elapsed_secs)

	def _is_nested_match(self, offset, length):
		if self._args.extract_nested:
			# Completely disregard if we've already captured this.
			return False
		interval = Interval.begin_length(offset, length)
		if self._matches.fully_contained_in_subinterval(interval):
			# We already have this match.
			return True
		else:
			self._matches.add(interval)
			return False

	def _is_known_blob(self, data):
		if self._args.allow_non_unique_blobs:
			# We record the exact same file twice, always.
			return False
		blob_hash = hashlib.sha256(data).digest()
		if blob_hash in self._hashes:
			return True
		else:
			self._hashes.add(blob_hash)
			return False

	def _record_finding(self, offset, data_type, extension, data, encode_pem_marker = None, orig_extension = None):
		if orig_extension is None:
			orig_extension = data_type

		if len(data) == 0:
			self._stats.record_finding(offset, len(data), data_type, orig_extension, "discard:zero_length")
			return

		if self._is_nested_match(offset, len(data)):
			self._stats.record_finding(offset, len(data), data_type, orig_extension, "discard:nested")
			self._log.debug("Found %s/%s at offset 0x%x, length %d bytes, not recording nested match.", data_type, orig_extension, offset, len(data))
			return

		if self._is_known_blob(data):
			self._stats.record_finding(offset, len(data), data_type, orig_extension, "discard:non-unique")
			self._log.debug("Found %s/%s at offset 0x%x, length %d bytes, not recording non-unique match.", data_type, orig_extension, offset, len(data))
			return

		filename_args = {
			"otype":	orig_extension,
			"type":		data_type,
			"offset":	offset,
			"ext":		extension,
		}
		filename = self._args.outdir + "/" + (self._args.outmask % filename_args)
		self._stats.record_finding(offset, len(data), data_type, orig_extension, "written", filename)
		self._log.info("Found %s/%s at offset 0x%x, length %d bytes, saved as %s", data_type, orig_extension, offset, len(data), filename)

		if encode_pem_marker is not None:
			output_data = (PEMDataTools.data2pem(data, encode_pem_marker) + "\n").encode()
		else:
			output_data = data
		with open(filename, "wb") as f:
			f.write(output_data)

	def _find_pem(self, offset, data):
		self._stats.pem_potential_match()
		textdata = data.decode("ascii", errors = "ignore")
		result = self._PEM_BEGIN.match(textdata)
		if result is None:
			return
		result = result.groupdict()
		marker = result["marker"]
		full_re = re.compile("-----BEGIN %s-----(?P<pem_data>.*?)-----END %s-----" % (marker, marker), flags = re.DOTALL | re.MULTILINE)
		result = full_re.match(textdata)
		if result is None:
			return
		result = result.groupdict()
		pem_data = result["pem_data"]
		pem_data = pem_data.replace("\r", "")
		pem_data = pem_data.replace("\n", "")
		pem_data = pem_data.replace("\t", "")
		pem_data = pem_data.replace(" ", "")
		der_data = base64.b64decode(pem_data)

		self._stats.pem_successful_decode()
		data_type = self._MARKERS.get(marker, "unknown")
		self._record_finding(offset = offset, data_type = data_type, extension = "pem", data = der_data, encode_pem_marker = marker)

	def _find_der(self, offset, data):
		self._stats.der_potential_match()

		for der_candidate in self._active_der_types:
			try:
				self._stats.der_attempt_decode()
				(asn1, tail) = pyasn1.codec.der.decoder.decode(data, asn1Spec = der_candidate.asn1_spec)
				if len(tail) == 0:
					asn1_data = data
				else:
					asn1_data = data[:-len(tail)]
				self._stats.der_successful_decode()

				if (not self._args.disable_der_sanity_checks) and (der_candidate.sanity_check_fn is not None):
					# We want sanity checks enabled and for the successfully
					# deserialized ASN.1 blob there is a handler registered.
					# Execute it (it'll throw an exception on failure, which
					# we'll catch).
					der_candidate.sanity_check_fn(asn1)
				self._stats.der_passed_plausibility()

				if self._args.keep_original_der or (der_candidate.pem_marker is None):
					# Should not or cannot re-encode as PEM, write DER file
					self._record_finding(offset = offset, data_type = der_candidate.data_type, extension = der_candidate.extension, data = asn1_data)
				else:
					self._record_finding(offset = offset, data_type = der_candidate.data_type, extension = "pem", data = asn1_data, encode_pem_marker = der_candidate.pem_marker, orig_extension = der_candidate.extension)

			except pyasn1.error.PyAsn1Error as e:
				pass
			except _DERSanityCheck.SanityCheckFailedException as e:
				self._log.debug("Potential %s blob encountered at offset 0x%x, but failed sanity check: %s", der_candidate.data_type, offset, str(e))
				self._stats.der_failed_plausibility()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2019 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

class ConsolePrinter():
	def __init__(self, fp = None):
		self._subs = [ ]
		self._fp = fp

	def add_sub(self, text, replacement):
		self._subs.append((text, replacement))
		return self

	def add_subs(self, substitutions):
		for (text, replacement) in substitutions.items():
			self.add_sub(text, replacement)
		return self

	def sub(self, line):
		for (text, replacement) in self._subs:
			line = line.replace(text, replacement)
		return line

	def heading(self, line = ""):
		print(line, file = self._fp)
		print((len(line) * "-"), file = self._fp)

	def print(self, line = ""):
		print(self.sub(line), file = self._fp)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
import datetime
from x509sak.tests import BaseTest
from x509sak.WorkDir import WorkDir
from x509sak.SubprocessExecutor import SubprocessExecutor
from x509sak.CertificateRevocationList import CertificateRevocationList

class CmdLineTestsCreateCRL(BaseTest):
	def test_create_crl(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "root_ca" ])
			self._run_x509sak([ "createcrl", "root_ca", "output.crl" ])
			crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-text", "-noout" ]).run().stdout
			self.assertIn(b"No Revoked Certificates.", crl)

			der_crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-outform", "der" ]).run().stdout
			crl = CertificateRevocationList(der_crl)
			self.assertEqual(crl.crt_count, 0)

	def test_revoke_crt_crl(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "root_ca" ])
			self._run_x509sak([ "createcrt", "-c", "root_ca", "client.key", "client.crt" ])

			self._run_x509sak([ "revokecrt", "root_ca", "client.crt" ])
			self._run_x509sak([ "createcrl", "root_ca", "output.crl" ])
			crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-text", "-noout" ]).run().stdout
			self.assertIn(b"Revoked Certificates:", crl)

			der_crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-outform", "der" ]).run().stdout
			crl = CertificateRevocationList(der_crl)
			self.assertEqual(crl.crt_count, 1)

	def test_crl_hash(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "root_ca" ])
			self._run_x509sak([ "createcrl", "root_ca", "--hashfnc", "sha256", "output.crl" ])
			crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-text", "-noout" ]).run().stdout
			self.assertIn(b"ecdsa-with-SHA256", crl)

			self._run_x509sak([ "createcrl", "root_ca", "--hashfnc", "sha384", "output.crl" ])
			crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-text", "-noout" ]).run().stdout
			self.assertIn(b"ecdsa-with-SHA384", crl)

	def test_crl_validity(self):
		with tempfile.TemporaryDirectory() as tempdir, WorkDir(tempdir):
			self._run_x509sak([ "createca", "root_ca" ])

			for days in [ 1, 30 ]:
				self._run_x509sak([ "createcrl", "root_ca", "-d", str(days), "output.crl" ])
				now = datetime.datetime.utcnow()
				der_crl = SubprocessExecutor([ "openssl", "crl", "-in", "output.crl", "-outform", "der" ]).run().stdout
				crl = CertificateRevocationList(der_crl)

				self.assertAlmostEqual((crl.this_update - now).total_seconds(), 0, delta = 60)
				self.assertAlmostEqual((crl.next_update - now).total_seconds(), 86400 * days, delta = 60)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import re
import collections
import json
import string
import functools
from x509sak.Tools import JSONTools
from x509sak.Exceptions import InvalidInternalDataException

@functools.total_ordering
class RichJudgementCode():
	def __init__(self, name, description, **kwargs):
		self._name = name
		self._description = description
		self._topic = kwargs.get("topic")
		self._flags = kwargs.get("flags")
		self._order = kwargs.get("order")

	@property
	def order(self):
		return self._order

	@property
	def name(self):
		return self._name

	@property
	def description(self):
		return self._description

	@property
	def topic(self):
		return self._topic

	@classmethod
	def from_node(cls, node):
		return cls(name = node.long_id, description = node.attrs["desc"], topic = node.attrs.get("topic"), order = node.order)

	@property
	def cmptuple(self):
		return (self.order, self.name)

	def __hash__(self):
		return hash(self.cmptuple)

	def __eq__(self, other):
		return self.cmptuple == other.cmptuple

	def __lt__(self, other):
		return self.cmptuple < other.cmptuple

	def __repr__(self):
		return "RichJudgementCode<%s>" % (self.name)

class StructureNode():
	_IMPORT_REGEX = re.compile(r"(?P<export_root_point>\+)?(?P<name>[a-zA-Z0-9_]+)(?P<import_contents>/\*)?(:(?P<flags>[a-zA-Z0-9_,]+))?({(?P<substitutions>[^}]+)})?")
	_ImportStatement = collections.namedtuple("ImportStatement", [ "name", "import_contents", "export_root_point", "flags", "substitutions" ])
	_LABEL_REGULAR_CHARS = set(string.ascii_lowercase + string.ascii_uppercase + string.digits)
	_LABEL_UNDERSCORE_CHARS = set("/")
	_ALLOWED_ATTRIBUTES = set([ "short_id", "long_id", "import", "export", "flags", "desc", "label", "require", "topic", "clone_of" ])
	_ALLOWED_FLAGS = set([ "datapoint" ])

	def __init__(self, name, children = None, attributes = None):
		self._name = name
		self._children = children if (children is not None) else [ ]
		self._attributes = attributes if (attributes is not None) else { }
		self._imported = collections.defaultdict(list)
		self._order = None
		assert(isinstance(self._children, list))
		assert(isinstance(self._attributes, dict))
		assert(all(key in self._ALLOWED_ATTRIBUTES for key in self._attributes))
		if "flags" in self._attributes:
			self._attributes["flags"] = frozenset(self._attributes["flags"])
			assert(all(flag in self._ALLOWED_FLAGS for flag in self._attributes["flags"]))
		if "import" in self._attributes:
			if isinstance(self._attributes["import"], str):
				self._attributes["import"] = [ self._attributes["import"] ]
			self._attributes["import"] = tuple(self.parse_import(import_str) for import_str in self._attributes["import"])

	def clone(self, filter_predicate = None, memorize_source = False):
		if not self.satisfies(filter_predicate):
			return None
		children = [ child.clone(filter_predicate = filter_predicate, memorize_source = memorize_source) for child in self.children if child.satisfies(filter_predicate) ]
		attributes = dict(self.attrs)
		if memorize_source:
			attributes["clone_of"] = self.long_id
		return StructureNode(name = self.name, children = children, attributes = attributes)

	def filter(self, filter_predicate):
		self._children = [ child.filter(filter_predicate = filter_predicate) for child in self.children if child.satisfies(filter_predicate) ]
		return self

	@classmethod
	def parse_import(cls, import_str):
		if isinstance(import_str, cls._ImportStatement):
			return import_str
		result = cls._IMPORT_REGEX.fullmatch(import_str)
		if not result:
			raise Exception("Not a valid import: '%s'" % (import_str))
		result = result.groupdict()

		attributes = {
			"name": result["name"],
			"import_contents": result["import_contents"] is not None,
			"export_root_point": result["export_root_point"] is not None,
			"flags": frozenset(result["flags"].split(",")) if (result["flags"] is not None) else tuple(),
			"substitutions": dict(),
		}
		if result["substitutions"] is not None:
			keyvalues = result["substitutions"].split(",")
			keyvalues = [ keyvalue.split("=", maxsplit = 1) for keyvalue in keyvalues ]
			attributes["substitutions"] = dict(keyvalues)
		return cls._ImportStatement(**attributes)

	@property
	def name(self):
		return self._name

	@property
	def is_leaf(self):
		return len(self._children) == 0

	@property
	def order(self):
		return self._order

	@order.setter
	def order(self, value):
		self._order = value

	@property
	def children(self):
		return iter(self._children)

	@property
	def children_leaf(self):
		return (child for child in self._children if (child.is_leaf))

	@property
	def children_nonleaf(self):
		return (child for child in self._children if (not child.is_leaf))

	@classmethod
	def transform_label_case(cls, text):
		next_uppercase = True
		label = [ ]
		for char in text:
			if char == " ":
				next_uppercase = True
			elif char in cls._LABEL_REGULAR_CHARS:
				if next_uppercase:
					char = char.upper()
					next_uppercase = False
				label.append(char)
			elif char in cls._LABEL_UNDERSCORE_CHARS:
				label.append("_")
		return "".join(label)

	@property
	def label_id(self):
		if self.name is None:
			# Root node has no label
			return None
		else:
			if self.attrs.get("label") is None:
				input_name = self.name
			else:
				input_name = self.attrs["label"]
			return self.transform_label_case(input_name)

	@property
	def long_id(self):
		return self.attrs.get("long_id")

	@property
	def attrs(self):
		return self._attributes

	def has_attribute(self, key):
		return key in self._attributes

	def add_import(self, import_source_name, imports):
		self._imported[import_source_name].append(imports)

	def satisfies(self, filter_predicate):
		if filter_predicate is None:
			return True
		return filter_predicate(self)

	def propagate_attribute_subtree(self, key):
		return self.apply_attribute_subtree(key, self.attrs.get(key))

	def apply_attribute_subtree(self, key, value):
		current_value = self.attrs.get(key)
		if current_value is not None:
			new_value = current_value
		else:
			new_value = value
		if new_value is not None:
			self.attrs[key] = new_value
		for child in self.children:
			child.apply_attribute_subtree(key, new_value)

	def purge_attribute(self, key):
		if key in self.attrs:
			del self.attrs[key]

	def purge_attribute_recursively(self, key):
		self.purge_attribute(key)
		for child in self.children:
			child.purge_attribute_recursively(key)

	def append_child(self, node):
		assert(isinstance(node, StructureNode))
		self._children.append(node)

	def append_children_of(self, node):
		for child in node.children:
			self.append_child(child)

	@classmethod
	def parse(cls, key, values):
		children = [ cls.parse(child_key, child_values) for (child_key, child_values) in values.items() if not child_key.startswith("_") ]
		attributes = { attribute_name[1:]: attribute_value for (attribute_name, attribute_value) in values.items() if (attribute_name.startswith("_")) and (not attribute_name.startswith("__")) }
		return cls(name = key, children = children, attributes = attributes)

	def dump(self, text = None, indent = 0):
		if text is not None:
			print(("=" * 60) + " " + text + " " + ("=" * 60))
		indent_str = ("   " * indent)
		if len(self._attributes) == 0:
			attrstr = ""
		else:
			attrstr = " %s" % (str(self.attrs))
		print("%s%s%s" % (indent_str, self.name, attrstr))
		for child in self.children:
			child.dump(indent = indent + 1)
		if text is not None:
			print("~" * 120)
			print()

	def walk_ordered(self, callback):
		callback(self)
		for child in sorted(self.children_leaf, key = lambda node: node.long_id):
			child.walk_ordered(callback)
		for child in sorted(self.children_nonleaf, key = lambda node: node.long_id):
			child.walk_ordered(callback)

	def walk(self, callback):
		callback(self)
		for child in self.children:
			child.walk(callback)

	def assign_nodes_long_ids(self, prefix = None):
		if self.label_id is not None:
			if prefix is None:
				long_id = self.label_id
			else:
				long_id = prefix + "_" + self.label_id
			self._attributes["long_id"] = long_id
		else:
			long_id = None
		for child in self._children:
			child.assign_nodes_long_ids(prefix = long_id)

	def __repr__(self):
		if self.has_attribute("desc"):
			return "Node<%s (%s)>" % (self.name, self.attrs["desc"])
		else:
			return "Node<%s>" % (self.name)

class JudgementStructure():
	def __init__(self, structure_data, verbose = False):
		self._verbose = verbose
		self._root = StructureNode.parse(None, structure_data)
		self._root.propagate_attribute_subtree("export")
		self._nodes_by_short_id = self._find_nodes_by_short_id()
		self._root.assign_nodes_long_ids()

		self._process_imports()
		self._root = self._root.filter(filter_predicate = lambda node: not node.attrs.get("export"))
		self._root.assign_nodes_long_ids()
		self._root.propagate_attribute_subtree("topic")
		self._assign_order()

	@property
	def root(self):
		return self._root

	def _find_nodes_by_short_id(self):
		short_ids = { }
		def visit(node):
			short_id = node.attrs.get("short_id")
			if short_id is not None:
				if short_id in short_ids:
					raise Exception("Duplicate short ID: %s" % (short_id))
				short_ids[short_id] = node
		self._root.walk(visit)
		return short_ids

	def _process_imports(self):
		def do_import(target, source, import_statement):
			def filter_predicate(node):
				requirements = [
					node.attrs.get("export") is True,
				]
				if node.has_attribute("require"):
					require = node.attrs["require"]
					if not require.startswith("!"):
						requirements.append(require in import_statement.flags)
					else:
						requirements.append(require[1:] not in import_statement.flags)
				return all(requirements)

			def apply_substitutions(node):
				if node.has_attribute("desc"):
					desc = node.attrs["desc"]
					for (search, replace) in import_statement.substitutions.items():
						desc = desc.replace("$" + search, replace)
					node.attrs["desc"] = desc

			source_clone = source.clone(filter_predicate = filter_predicate, memorize_source = True)
			if source_clone is None:
				raise Exception("Cannot import node that is not exported.")
			if not target.attrs.get("export"):
				source_clone.purge_attribute_recursively("export")
			source_clone.purge_attribute_recursively("short_id")
			source_clone.purge_attribute_recursively("require")
			if len(import_statement.substitutions) > 0:
				source_clone.walk(apply_substitutions)
			target.purge_attribute("import")
			if import_statement.export_root_point:
				if import_statement.import_contents:
					for child in source_clone.children:
						target.add_import(source.label_id, child)
				else:
					target.add_import(source.label_id, source_clone)

			if import_statement.import_contents:
				target.append_children_of(source_clone)
			else:
				target.append_child(source_clone)

		def visit(node):
			import_statements = node.attrs.get("import")
			if import_statements is None:
				return

			for import_statement in import_statements:
				if import_statement.name not in self._nodes_by_short_id:
					raise Exception("Import of '%s' requested, but no node by that short ID found." % (import_statement.name))
				import_node = self._nodes_by_short_id[import_statement.name]

				if self._verbose:
					print("%s imports %s" % (node, import_node))

				# Recursively descent before importing, in case the import imports something of its own
				visit(import_node)

				# Then import
				do_import(node, import_node, import_statement)

				if self._verbose:
					node.dump()
					print()

		self._root.walk(visit)

	def _assign_order(self):
		order_no = [ 0 ]
		def visit(node):
			order_no[0] += 1
			node.order = order_no[0]
		self._root.walk_ordered(visit)

	def _create_enum_class(self):
		exported_nodes = { }
		def visit(node):
			if node.attrs.get("desc") is not None:
				if node.long_id in exported_nodes:
					raise InvalidInternalDataException("Note with long ID '%s' appears twice (%s and %s)." % (node.long_id, exported_nodes[node.long_id], node))
				exported_nodes[node.long_id] = node
		self._root.walk(visit)

		exported_nodes = { name: RichJudgementCode.from_node(node) for (name, node) in exported_nodes.items() }
		enum_class = enum.Enum("JudgementCode", exported_nodes)
		return enum_class

	def _create_inheritance_dict(self):
		inheritance_dict = { }

		def create_cloned_node_dict(cloned_subtree):
			cloned_nodes = { }
			def visit(node):
				orig_name = node.attrs["clone_of"]
				node_name = node.long_id
				cloned_nodes[orig_name] = node_name
			cloned_subtree.walk(visit)
			return cloned_nodes

		def visit(node):
			if len(node._imported) > 0:
				inheritance_dict[node.long_id] = { }
				for (import_name, cloned_subtrees) in node._imported.items():
					for cloned_subtree in cloned_subtrees:
						inheritance_dict[node.long_id].update(create_cloned_node_dict(cloned_subtree))

		self._root.walk(visit)
		return inheritance_dict

	def create_extended_enum_class(self):
		judgement_code_class = self._create_enum_class()
		inheritance = self._create_inheritance_dict()

		resolved_inheritance = { }
		for (root_point, imports) in inheritance.items():
			resolved_inheritance[root_point] = { }
			for (general_name, inherited_name) in imports.items():
				try:
					resolved_inheritance[root_point][general_name] = getattr(judgement_code_class, inherited_name)
				except AttributeError:
					# Intermediate node with no content (e.g.,
					# "Encoding_ASN1DER_Structures_DistinguishedName_RDN")
					pass

		judgement_code_class.inheritance = resolved_inheritance
		return judgement_code_class

	@classmethod
	def load_from_json(cls, filename):
		with open(filename) as f:
			structure_data = json.load(f)
		return cls(structure_data)

def create_judgement_structure(verbose = False):
	structure_data = { }
	for structure_json_name in [ "number_theoretic.json", "encoding.json", "cryptography.json", "x509ext.json", "x509cert.json" ]:
		partial_data = JSONTools.load_internal("x509sak.data.judgements", structure_json_name)
		structure_data.update(partial_data)
	structure = JudgementStructure(structure_data, verbose = verbose)
	return structure

def create_judgement_code_class(verbose = False):
	structure = create_judgement_structure(verbose = verbose)
	return structure.create_extended_enum_class()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2021 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import os
import sys
import logging
import traceback
import x509sak
from x509sak.actions.ActionBuildChain import ActionBuildChain
from x509sak.actions.ActionGraphPool import ActionGraphPool
from x509sak.actions.ActionFindCert import ActionFindCert
from x509sak.actions.ActionCreateCA import ActionCreateCA
from x509sak.actions.ActionCreateCSR import ActionCreateCSR
from x509sak.actions.ActionSignCSR import ActionSignCSR
from x509sak.actions.ActionRevokeCRT import ActionRevokeCRT
from x509sak.actions.ActionCreateCRL import ActionCreateCRL
from x509sak.actions.ActionGenerateBrokenDSA import ActionGenerateBrokenDSA
from x509sak.actions.ActionGenerateBrokenRSA import ActionGenerateBrokenRSA
from x509sak.actions.ActionDumpKey import ActionDumpKey
from x509sak.actions.ActionExamineCert import ActionExamineCert
from x509sak.actions.ActionForgeCert import ActionForgeCert
from x509sak.actions.ActionScrape import ActionScrape
from x509sak.actions.ActionHashPart import ActionHashPart
from x509sak.actions.ActionTLSClient import ActionTLSClient
from x509sak.actions.ActionTLSParse import ActionTLSParse
from x509sak.actions.ActionDebug import ActionDebug
from x509sak.actions.ActionJudgementCode import ActionJudgementCode
try:
	from x509sak.actions.ActionTestcaseGen import ActionTestcaseGen
except ImportError:
	ActionTestcaseGen = None
from x509sak.CmdLineArgs import KeyValue
from x509sak.KeySpecification import KeySpecification
from x509sak.Exceptions import UserErrorException, InvisibleUserErrorException, CmdExecutionFailedException
from x509sak.SubprocessExecutor import SubprocessExecutor
from .FriendlyArgumentParser import baseint, baseint_unit
from .MultiCommand import MultiCommand

def main():

	_default_so_search_path = "/usr/local/lib:/usr/lib:/usr/lib/x86_64-linux-gnu:/usr/lib/x86_64-linux-gnu/openssl-1.0.2/engines:/usr/lib/x86_64-linux-gnu/engines-1.1"

	if "X509SAK_VERBOSE_EXECUTION" in os.environ:
		SubprocessExecutor.set_all_verbose()
	if "X509SAK_PAUSE_FAILED_EXECUTION" in os.environ:
		SubprocessExecutor.pause_after_failed_execution()
	if "X509SAK_PAUSE_BEFORE_EXECUTION" in os.environ:
		SubprocessExecutor.pause_before_execution()

	mc = MultiCommand(trailing_text = "version: x509sak v%s" % (x509sak.VERSION))

	def genparser(parser):
		parser.add_argument("-s", "--ca-source", metavar = "path", action = "append", default = [ ], help = "CA file (PEM format) or directory (containing .pem/.crt files) to include when building the chain. Can be specified multiple times to include multiple locations.")
		parser.add_argument("--inform", choices = [ "pem", "der" ], default = "pem", help = "Specifies input file format for certificate. Possible options are %(choices)s. Default is %(default)s.")
		parser.add_argument("--order-leaf-to-root", action = "store_true", help = "By default, certificates are ordered with the root CA first and intermediate certificates following up to the leaf. When this option is specified, the order is inverted and go from leaf certificate to root.")
		parser.add_argument("--allow-partial-chain", action = "store_true", help = "When building the certificate chain, a full chain must be found or the chain building fails. When this option is specified, also partial chain matches are permitted, i.e., not going up to a root CA. Note that this can have undesired side effects when no root certificates are found at all (the partial chain will then consist of only the leaf certificate itself).")
		parser.add_argument("--dont-trust-crtfile", action = "store_true", help = "When there's multiple certificates in the given crtfile in PEM format, they're by default all added to the truststore. With this option, only the leaf cert is taken from the crtfile and they're not added to the trusted pool.")
		parser.add_argument("--outform", choices = [ "rootonly", "intermediates", "fullchain", "all-except-root", "multifile", "pkcs12" ], default = "fullchain", help = "Specifies what to write into the output file. Possible options are %(choices)s. Default is %(default)s. When specifying multifile, a %%d format must be included in the filename to serve as a template; typical printf-style formatting can be used of course (e.g., %%02d).")
		parser.add_argument("--private-key", metavar = "filename", type = str, help = "When creating a PKCS#12 output file, this private key can also be included. By default, only the certificates are exported.")
		parser.add_argument("--pkcs12-legacy-crypto", action = "store_true", help = "Use crappy crypto to encrypt a PKCS#12 exported private key.")
		group = parser.add_mutually_exclusive_group()
		group.add_argument("--pkcs12-no-passphrase", action = "store_true", help = "Do not use any passphrase to protect the PKCS#12 private key.")
		group.add_argument("--pkcs12-passphrase-file", metavar = "filename", type = str, help = "Read the PKCS#12 passphrase from the first line of the given file. If omitted, by default a random passphrase will be generated and printed on stderr.")
		parser.add_argument("-o", "--outfile", metavar = "file", help = "Specifies the output filename. Defaults to stdout.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("crtfile", metavar = "crtfile", type = str, help = "Certificate that a chain shall be build for, in PEM format.")
	mc.register("buildchain", "Build a certificate chain", genparser, action = ActionBuildChain, aliases = [ "bc" ])

	def genparser(parser):
		parser.add_argument("-c", "--color-scheme", choices = sorted(list(ActionGraphPool.get_supported_colorschemes())), default = "expiration", help = "Color scheme to use when coloring the certificates. Can either color by expiration date, by certificate type (client/server/CA/...), key type (RSA/ECC/etc), signature type (used hash function) or overall security level. Defaults to %(default)s.")
		parser.add_argument("--abbreviate-to", metavar = "charcnt", type = int, default = 30, help = "Abbreviate each line to this amount of characters. Defaults to %(default)d characters.")
		parser.add_argument("-l", "--label", metavar = "text", default = [ ], action = "append", help = "Label that is printed in the certificate nodes. Can be given multiple times to specify multiple lines. Substitutions that are supported are %s. Defaults to %s." % (", ".join(sorted(ActionGraphPool.get_supported_substitutions())), [ line.replace("%", "%%") for line in ActionGraphPool.get_default_label() ]))
		parser.add_argument("-f", "--format", choices = [ "dot", "png", "ps", "pdf" ], default = None, help = "Specifies the output file format. Can be one of %(choices)s. When unspecified, the file extension out the output file is used to determine the file type.")
		parser.add_argument("-o", "--outfile", metavar = "file", required = True, help = "Specifies the output filename. Mandatory argument.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("crtsource", metavar = "crtsource", nargs = "+", type = str, help = "Certificate file (in PEM format) or directory (containting PEM-formatted .pem or .crt files) which should be included in the graph.")
	mc.register("graph", "Graph a certificate pool", genparser, action = ActionGraphPool)

	def genparser(parser):
		parser.add_argument("-h", "--hashval", metavar = "hash", type = str, help = "Find only certificates with a particular hash prefix.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("crtsource", metavar = "crtsource", nargs = "+", type = str, help = "Certificate file (in PEM format) or directory (containting PEM-formatted .pem or .crt files) which should be included in the search.")
	mc.register("findcrt", "Find a specific certificate", genparser, action = ActionFindCert)

	def genparser(parser):
		group = parser.add_mutually_exclusive_group()
		group.add_argument("-g", "--gen-keyspec", metavar = "keyspec", type = KeySpecification.from_cmdline_str, help = "Private key specification to generate. Examples are rsa:1024 or ecc:secp256r1. Defaults to ecc:secp384r1.")
		group.add_argument("-w", "--hardware-key", metavar = "pkcs11uri", type = str, help = "Use a hardware token which stores the private key. The parameter gives the pkcs11 URI, e.g., 'pkcs11:object=mykey;type=private'")
		parser.add_argument("--pkcs11-so-search", metavar = "path", type = str, default = _default_so_search_path, help = "Gives the path that will be searched for the \"dynamic\" and \"module\" shared objects. The \"dynamic\" shared object is libpkcs11.so, the \"module\" shared object can be changed by the --pkcs11-module option. The search path defaults to %(default)s.")
		parser.add_argument("--pkcs11-module", metavar = "sofile", type = str, default = "opensc-pkcs11.so", help = "Name of the \"module\" shared object when using PKCS#11 keys. Defaults to %(default)s.")
		parser.add_argument("-p", "--parent-ca", metavar = "capath", type = str, help = "Parent CA directory. If omitted, CA certificate will be self-signed.")
		parser.add_argument("-s", "--subject-dn", metavar = "subject", type = str, default = "/CN=Root CA", help = "CA subject distinguished name. Defaults to %(default)s.")
		parser.add_argument("-d", "--validity-days", metavar = "days", type = int, default = 365, help = "Number of days that the newly created CA will be valid for. Defaults to %(default)s days.")
		parser.add_argument("-h", "--hashfnc", metavar = "alg", type = str, default = "sha384", help = "Hash function to use for signing the CA certificate. Defaults to %(default)s.")
		parser.add_argument("--serial", metavar = "serial", type = baseint, help = "Serial number to use for root CA certificate. Randomized by default.")
		parser.add_argument("--allow-duplicate-subjects", action = "store_true", help = "By default, subject distinguished names of all valid certificates below one CA must be unique. This option allows the CA to have duplicate distinguished names for certificate subjects.")
		parser.add_argument("--extension", metavar = "key=value", type = KeyValue, action = "append", default = [ ], help = "Additional certificate X.509 extension to include on top of the default CA extensions. Can be specified multiple times.")
		parser.add_argument("-f", "--force", action = "store_true", help = "By default, the capath will not be overwritten if it already exists. When this option is specified the complete directory will be erased before creating the new CA.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("capath", metavar = "capath", type = str, help = "Directory to create the new CA in.")
	mc.register("createca", "Create a new certificate authority (CA)", genparser, aliases = [ "genca" ], action = ActionCreateCA)

	def genparser(parser):
		parser.add_argument("-g", "--gen-keyspec", metavar = "keyspec", type = KeySpecification.from_cmdline_str, help = "Private key specification to generate for the certificate or CSR when it doesn't exist. Examples are rsa:1024 or ecc:secp256r1.")
		parser.add_argument("-k", "--keytype", choices = [ "pem", "der", "hw" ], default = "pem", help = "Private key type. Can be any of %(choices)s. Defaults to %(default)s.")
		parser.add_argument("-s", "--subject-dn", metavar = "subject", type = str, default = "/CN=New Cert", help = "Certificate/CSR subject distinguished name. Defaults to %(default)s.")
		parser.add_argument("-d", "--validity-days", metavar = "days", type = int, default = 365, help = "When creating a certificate, number of days that the certificate will be valid for. Defaults to %(default)s days.")
		parser.add_argument("-h", "--hashfnc", metavar = "alg", type = str, default = None, help = "Hash function to use for signing when creating a certificate. Defaults to the default hash function specified in the CA config.")
		parser.add_argument("-t", "--template", choices = [ "rootca", "ca", "tls-server", "tls-client" ], help = "Template to use for determining X.509 certificate extensions. Can be one of %(choices)s. By default, no extensions are included except for SAN.")
		parser.add_argument("--san-dns", metavar = "FQDN", type = str, action = "append", default = [ ], help = "Subject Alternative DNS name to include in the certificate or CSR. Can be specified multiple times.")
		parser.add_argument("--san-ip", metavar = "IP", type = str, action = "append", default = [ ], help = "Subject Alternative IP address to include in the certificate or CSR. Can be specified multiple times.")
		parser.add_argument("--extension", metavar = "key=value", type = KeyValue, action = "append", default = [ ], help = "Additional certificate X.509 extension to include on top of the extensions in the template and by the SAN parameters. Can be specified multiple times.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite the output file if it already exists.")
		parser.add_argument("-c", "--create-crt", metavar = "capath", help = "Instead of creating a certificate signing request, directly create a certificate instead. Needs to supply the CA path that should issue the certificate.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("key_filename", metavar = "in_key_filename", type = str, help = "Filename of the input private key or PKCS#11 URI (as specified in RFC7512 in case of a hardware key type.")
		parser.add_argument("out_filename", metavar = "out_filename", type = str, help = "Filename of the output certificate signing request or certificate.")
	mc.register("createcsr", "Create a new certificate signing request (CSR) or certificate", genparser, aliases = [ "gencsr", "createcrt", "gencrt" ], action = ActionCreateCSR)

	def genparser(parser):
		parser.add_argument("-s", "--subject-dn", metavar = "subject", type = str, help = "Certificate's subject distinguished name. Defaults to the subject given in the CSR.")
		parser.add_argument("-d", "--validity-days", metavar = "days", type = int, default = 365, help = "Number of days that the newly created certificate will be valid for. Defaults to %(default)s days.")
		parser.add_argument("-h", "--hashfnc", metavar = "alg", type = str, default = None, help = "Hash function to use for signing. Defaults to the default hash function specified in the CA config.")
		parser.add_argument("-t", "--template", choices = [ "rootca", "ca", "tls-server", "tls-client" ], help = "Template to use for determining X.509 certificate extensions. Can be one of %(choices)s. By default, no extensions are included except for SAN.")
		parser.add_argument("--san-dns", metavar = "FQDN", type = str, action = "append", default = [ ], help = "Subject Alternative DNS name to include in the certificate. Can be specified multiple times.")
		parser.add_argument("--san-ip", metavar = "IP", type = str, action = "append", default = [ ], help = "Subject Alternative IP address to include in the CRT. Can be specified multiple times.")
		parser.add_argument("--extension", metavar = "key=value", type = KeyValue, action = "append", default = [ ], help = "Additional certificate X.509 extension to include on top of the extensions in the template and by the SAN parameters. Can be specified multiple times.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite the output certificate file if it already exists.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("capath", metavar = "capath", type = str, help = "Directory of the signing CA.")
		parser.add_argument("csr_filename", metavar = "in_csr_filename", type = str, help = "Filename of the input certificate signing request.")
		parser.add_argument("crt_filename", metavar = "out_crt_filename", type = str, help = "Filename of the output certificate.")
	mc.register("signcsr", "Make a certificate authority (CA) sign a certificate signing request (CSR) and output the certificate", genparser, action = ActionSignCSR)

	def genparser(parser):
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("capath", metavar = "capath", type = str, help = "CA which created the certificate.")
		parser.add_argument("crt_filename", metavar = "crt_filename", type = str, help = "Filename of the output certificate.")
	mc.register("revokecrt", "Revoke a specific certificate", genparser, action = ActionRevokeCRT)

	def genparser(parser):
		parser.add_argument("-d", "--validity-days", metavar = "days", type = int, default = 30, help = "Number of days until the CRLs 'nextUpdate' field will expire. Defaults to %(default)s days.")
		parser.add_argument("-h", "--hashfnc", metavar = "alg", type = str, default = "sha256", help = "Hash function to use for signing the CRL. Defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("capath", metavar = "capath", type = str, help = "CA which should generate the CRL.")
		parser.add_argument("crl_filename", metavar = "crl_filename", type = str, help = "Filename of the output CRL.")
	mc.register("createcrl", "Generate a certificate revocation list (CRL)", genparser, action = ActionCreateCRL, aliases = [ "gencrl" ])

	def genparser(parser):
		parser.add_argument("-d", "--prime-db", metavar = "path", type = str, default = ".", help = "Prime database directory. Defaults to %(default)s and searches for files called primes_{bitlen}.txt in this directory.")
		parser.add_argument("-b", "--bitlen", metavar = "bits", type = int, default = 2048, help = "Bitlength of modulus. Defaults to %(default)d bits.")
		parser.add_argument("-e", "--public-exponent", metavar = "exp", type = baseint, default = 0x10001, help = "Public exponent e (or d in case --switch-e-d is specified) to use. Defaults to 0x%(default)x. Will be randomly chosen from 2..n-1 if set to -1.")
		parser.add_argument("--switch-e-d", action = "store_true", help = "Switch e with d when generating keypair.")
		parser.add_argument("--accept-unusable-key", action = "store_true", help = "Disregard integral checks, such as if gcd(e, phi(n)) == 1 before inverting e. Might lead to an unusable key or might fail altogether.")
		parser.add_argument("--carmichael-totient", action = "store_true", help = "By default, d is computed as the modular inverse of e to phi(n), the Euler Totient function. This computes d as the modular inverse of e to lambda(n), the Carmichael Totient function, instead.")
		parser.add_argument("--generator", metavar = "file", help = "When prime database is exhausted, will call the prime generator program as a subprocess to generate new primes. Otherwise, and the default behavior, is to fail.")
		group = parser.add_mutually_exclusive_group()
		group.add_argument("--gcd-n-phi-n", action = "store_true", help = "Generate a keypair in which gcd(n, phi(n)) != 1 by specially constructing the prime q. This will lead to a size disparity of p and q and requires 3-msb primes as input.")
		group.add_argument("--close-q", action = "store_true", help = "Use a value for q that is very close to the value of p so that search starting from sqrt(n) is computationally feasible to factor the modulus. Note that for this, the bitlength of the modulus must be evenly divisible by two.")
		parser.add_argument("--q-stepping", metavar = "int", type = baseint, default = 1, help = "When creating a close-q RSA keypair, q is chosen by taking p and incrementing it repeatedly by a random int from 2 to (2 * q-stepping). The larger q-stepping is therefore chosen, the further apart p and q will be. By default, q-stepping is the minimum value of %(default)d.")
		parser.add_argument("-o", "--outfile", metavar = "file", type = str, default = "broken_rsa.key", help = "Output filename. Defaults to %(default)s.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite output file if it already exists instead of bailing out.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
	mc.register("genbrokenrsa", "Generate broken RSA keys for use in penetration testing", genparser, action = ActionGenerateBrokenRSA)

	def genparser(parser):
		parser.add_argument("-d", "--prime-db", metavar = "path", type = str, default = ".", help = "Prime database directory. Defaults to %(default)s and searches for files called primes_{bitlen}.txt in this directory.")
		parser.add_argument("--generator", metavar = "file", help = "When prime database is exhausted, will call the prime generator program as a subprocess to generate new primes. Otherwise, and the default behavior, is to fail.")
		parser.add_argument("-o", "--outfile", metavar = "file", type = str, default = "broken_dsa.key", help = "Output filename. Defaults to %(default)s.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite output file if it already exists instead of bailing out.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("L_bits", metavar = "L_bits", type = int, help = "Bitlength of the modulus p, also known as L.")
		parser.add_argument("N_bits", metavar = "N_bits", type = int, help = "Bitlength of q, also known as N.")
	mc.register("genbrokendsa", "Generate broken DSA parameters for use in penetration testing", genparser, action = ActionGenerateBrokenDSA)

	def genparser(parser):
		parser.add_argument("-t", "--key-type", choices = [ "rsa", "ecc", "eddsa" ], default = "rsa", help = "Type of private key to import. Can be one of %(choices)s, defaults to %(default)s. Disregarded for public keys and determined automatically.")
		parser.add_argument("-p", "--public-key", action = "store_true", help = "Input is a public key, not a private key.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("key_filename", metavar = "key_filename", type = str, help = "Filename of the input key file in PEM format.")
	mc.register("dumpkey", "Dump a key in text form", genparser, action = ActionDumpKey)

	def genparser(parser):
		parser.add_argument("-p", "--purpose", choices = [ "ca", "tls-server", "tls-client" ], action = "append", default = [ ], help = "Check if the certificate is fit for the given purpose. Can be any of %(choices)s, can be specified multiple times.")
		parser.add_argument("-n", "--server-name", metavar = "fqdn", type = str, help = "Check if the certificate is valid for the given hostname.")
		parser.add_argument("-f", "--out-format", choices = [ "ansitext", "text", "json" ], default = "ansitext", help = "Determine the output format. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-i", "--in-format", choices = [ "pemcrt", "dercrt", "json", "host" ], default = "pemcrt", help = "Specifies the type of file that is read in. Can be either certificate files in PEM or DER format, a pre-processed JSON output from a previous run or a hostname[:port] combination to query a TLS server directly (port defaults to 443 if omitted). Valid choices are %(choices)s, defaults to %(default)s.")
		parser.add_argument("-r", "--parent-certificate", metavar = "pemfile", type = str, help = "Specifies a parent CA certificate that is used to run additional checks against the certificate.")
		parser.add_argument("--no-automatic-host-check", action = "store_true", help = "By default, when the input format is a given hostname, the server name is assumed as well and the purpose is assumed to be a TLS server. When this option is specified, these automatic checks are omitted.")
		parser.add_argument("--fast-rsa", action = "store_true", help = "Skip some time-intensive number theoretical tests for RSA moduli in order to speed up checking. Less thorough, but much faster.")
		parser.add_argument("--include-raw-data", action = "store_true", help = "Add the raw data such as base64-encoded certificate and signatures into the result as well.")
		parser.add_argument("--pretty-json", action = "store_true", help = "Prettyfy any generated JSON output.")
		parser.add_argument("-o", "--output", metavar = "filename", help = "Specify the output file. Defaults to stdout.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("infiles", metavar = "filename/uri", type = str, nargs = "+", help = "Filename of the input certificate or certificates in PEM format.")
	mc.register("examinecert", "Examine an X.509 certificate", genparser, action = ActionExamineCert, aliases = [ "analyze" ])

	def genparser(parser):
		parser.add_argument("--key_template", metavar = "path", default = "forged_%02d.key", help = "Output template for key files. Should contain '%%d' to indicate element in chain. Defaults to '%(default)s'.")
		parser.add_argument("--cert_template", metavar = "path", default = "forged_%02d.crt", help = "Output template for certificate files. Should contain '%%d' to indicate element in chain. Defaults to '%(default)s'.")
		parser.add_argument("-r", "--recalculate-keyids", action = "store_true", help = "By default, Subject Key Identifier and Authority Key Identifier X.509 extensions are kept as-is in the forged certificates. Specifying this will recalculate the IDs to fit the forged keys.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite key/certificate files.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("crt_filename", metavar = "crt_filename", type = str, help = "Filename of the input certificate or certificates PEM format.")
	mc.register("forgecert", "Forge an X.509 certificate", genparser, action = ActionForgeCert)

	def genparser(parser):
		parser.add_argument("--no-pem", action = "store_true", help = "Do not search for any PEM encoded blobs.")
		parser.add_argument("--no-der", action = "store_true", help = "Do not search for any DER encoded blobs.")
		parser.add_argument("-i", "--include-dertype", metavar = "class", action = "append", default = [ ], help = "Include the specified DER handler class in the search. Defaults to all known classes if omitted. Can be specified multiple times and must be one of %s." % (", ".join(sorted(ActionScrape.handler_classes))))
		parser.add_argument("-e", "--exclude-dertype", metavar = "class", action = "append", default = [ ], help = "Exclude the specified DER handler class in the search. Can be specified multiple times and must be one of %s." % (", ".join(sorted(ActionScrape.handler_classes))))
		parser.add_argument("--extract-nested", action = "store_true", help = "By default, fully overlapping blobs will not be extracted. For example, every X.509 certificate also contains a public key inside that would otherwise be found as well. When this option is given, any blobs are extracted regardless if they're fully contained in another blob or not.")
		parser.add_argument("--keep-original-der", action = "store_true", help = "When finding DER blobs, do not convert them to PEM format, but leave them as-is.")
		parser.add_argument("--allow-non-unique-blobs", action = "store_true", help = "For all matches, the SHA256 hash is used to determine if the data is unique and findings are by default only written to disk once. With this option, blobs that very likely are duplicates are written to disk for every occurrence.")
		parser.add_argument("--disable-der-sanity-checks", action = "store_true", help = "For DER serialization, not only is it checked that deserialization is possible, but additional checks are performed for some data types to ensure a low false-positive rate. For example, DSA signatures with short r/s pairs are discarded by default or implausible version numbers for EC keys. With this option, these sanity checks will be disabled and therefore structurally correct (but implausible) false-positives are also written.")
		parser.add_argument("--outmask", metavar = "mask", default = "scrape_%(offset)07x_%(type)s.%(ext)s", help = "Filename mask that's used for output. Defaults to %(default)s and can use printf-style substitutions offset, type and ext.")
		parser.add_argument("-w", "--write-json", metavar = "filename", type = str, help = "Write the stats with detailed information about matches into the given filename.")
		parser.add_argument("-o", "--outdir", metavar = "path", type = str, default = "scrape", help = "Output directory. Defaults to %(default)s.")
		parser.add_argument("-f", "--force", action = "store_true", help = "Overwrite key/certificate files and proceed even if outdir already exists.")
		parser.add_argument("-s", "--seek-offset", metavar = "offset", type = baseint_unit, default = "0", help = "Offset to seek into file. Supports hex/octal/binary prefixes and SI/binary SI (k, ki, M, Mi, etc.) suffixes. Defaults to %(default)s.")
		parser.add_argument("-l", "--analysis-length", metavar = "length", type = baseint_unit, default = None, help = "Amount of data to inspect at max. Supports hex/octal/binary prefixes and SI/binary SI (k, ki, M, Mi, etc.) suffixes. Defaults to everything until EOF is hit.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("filename", metavar = "filename", type = str, help = "File that should be scraped for certificates or keys.")
	mc.register("scrape", "Scrape input file for certificates, keys or signatures", genparser, action = ActionScrape)

	def genparser(parser):
		parser.add_argument("-h", "--hash-alg", metavar = "alg", choices = ActionHashPart.get_supported_hash_fncs() + [ "all" ], action = "append", default = [ ], help = "Hash function(s) that should be tried. Can be specified multiple times and defaults to all available hash functions. Can be any of %%(choices)s, but defaults to %s. Special value 'all' means all supported functions." % (", ".join(ActionHashPart.get_default_hash_fncs())))
		parser.add_argument("-o", "--seek-offset", metavar = "offset", type = baseint_unit, default = "0", help = "Offset to seek into file. Supports hex/octal/binary prefixes and SI/binary SI (k, ki, M, Mi, etc.) suffixes. Defaults to %(default)s.")
		parser.add_argument("--max-offset", metavar = "offset", type = baseint_unit, help = "Largest offset to consider. By default, this is end-of-file. Supports hex/octal/binary prefixes and SI/binary SI (k, ki, M, Mi, etc.) suffixes.")
		parser.add_argument("-a", "--variable-hash-length", metavar = "length", type = int, action = "append", default = [ ], help = "For hash functions which have a variable output length, try all of these hash lenghts. Length is given in bits and must be a multiple of 8. Can be supplied multiple times. Defaults to %s." % ((", ".join("%d" % (length) for length in ActionHashPart.get_default_variable_hash_lengths_bits()))))
		parser.add_argument("-l", "--analysis-length", metavar = "length", type = baseint_unit, default = None, help = "Amount of data to inspect at max. Supports hex/octal/binary prefixes and SI/binary SI (k, ki, M, Mi, etc.) suffixes. Defaults to everything until EOF is hit.")
		parser.add_argument("-s", "--search", metavar = "hexpattern", type = str, help = "Hexadecimal pattern that is expected in the hashing.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("filename", metavar = "filename", type = str, help = "File that should be hashed.")
	mc.register("hashpart", "Hash all substrings of a file and search for a particular hash value", genparser, action = ActionHashPart)

	def genparser(parser):
		parser.add_argument("-p", "--port", metavar = "port", default = 443, help = "Port to connect to. Defaults to %(default)d.")
		parser.add_argument("-s", "--starttls", choices = [ "smtp" ], help = "Perform a STARTTLS cleartext negotiation before sending TLS messages. Can be any of %(choices)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("servername", metavar = "filename", type = str, help = "Server name to connect to.")
	mc.register("tlsclient", "Act as a TLS client", genparser, action = ActionTLSClient, visible = False)

	def genparser(parser):
		parser.add_argument("-s", "--side", choices = [ "client", "server" ], default = "client", help = "Which side the messages are from, can be one of %(choices)s. Defaults to %(default)s.")
		parser.add_argument("-e", "--encoding", choices = [ "bin", "hex" ], default = "bin", help = "Encoding of the given file. Can be one of %(choices)s, defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("filename", metavar = "filename", type = str, help = "Binary file that contains the data stream.")
	mc.register("tlsparse", "Parse previously sniffed TLS messages and dump them", genparser, action = ActionTLSParse, visible = False)

	def genparser(parser):
		parser.add_argument("--der", action = "store_true", help = "Read in certificate in DER format instead of PEM.")
		parser.add_argument("-n", "--no-interact", action = "store_true", help = "Do not enter interactive console mode. Useful if you only want to execute specific commands and immediately quit.")
		parser.add_argument("-e", "--execute", metavar = "command", action = "append", default = [ ], help = "Execute the given command on the console. Can be specified multiple times.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
		parser.add_argument("crtfile", metavar = "filename", type = str, nargs = "*", help = "Certificate file that should be loaded in the console.")
	mc.register("debug", "Open an interactive Python console", genparser, aliases = [ "dbg" ], action = ActionDebug, visible = False)

	def genparser(parser):
		parser.add_argument("-a", "--action", choices = [ "list", "dump", "inherit" ], default = "list", help = "Specifies what information to print out, can be one of %(choices)s. Defaults to %(default)s.")
		parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
	mc.register("judgementcode", "Show information about judgement codes", genparser, aliases = [ "jc" ], action = ActionJudgementCode, visible = False)

	if ActionTestcaseGen is not None:
		def genparser(parser):
			parser.add_argument("-l", "--list-parameters", action = "store_true", help = "List possible parameters and values from the template.")
			parser.add_argument("-n", "--no-pem", action = "store_true", help = "Do not generate PEM data, just print the asciider")
			parser.add_argument("-o", "--output-dir", metavar = "path", type = str, default = "gen_tcs", help = "Output directory to store testcases in. Defaults to %(default)s.")
			parser.add_argument("-p", "--parameter", metavar = "key=value", type = str, action = "append", default = [ ], help = "Give a key/value parameter.")
			parser.add_argument("-v", "--verbose", action = "count", default = 0, help = "Increase verbosity level. Can be specified multiple times.")
			parser.add_argument("tcname", metavar = "basename", type = str, help = "Testcase name to render.")
		mc.register("testcasegen", "Generate testcase certificates", genparser, aliases = [ "gentc", "tcgen" ], action = ActionTestcaseGen, visible = False)

	try:
		mc.run(sys.argv[1:])
	except (UserErrorException, InvisibleUserErrorException) as e:
		if logging.root.level == logging.DEBUG:
			traceback.print_exc()
			print(file = sys.stderr)
		if isinstance(e, UserErrorException) or (logging.root.level == logging.DEBUG):
			print("%s: %s" % (e.__class__.__name__, str(e)), file = sys.stderr)
		else:
			print("Failure while processing this request: %s" % (e.__class__.__name__), file = sys.stderr)

		if isinstance(e, CmdExecutionFailedException):
			if len(e.execution_result.stderr) > 0:
				print("~" * 120)
				e.execution_result.dump()
		sys.exit(1)

if __name__ == "__main__":
	main()



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
import collections
from x509sak.OID import OIDDB

class LookupEnum(enum.Enum):
	@classmethod
	def lookup(cls, key, search_value):
		for item in cls:
			value_or_values = getattr(item.value, key)
			if isinstance(value_or_values, (list, tuple)):
				if any(member_value == search_value for member_value in value_or_values):
					return item
			else:
				if value_or_values == search_value:
					return item
		return None

def _to_enum(enum_name, items):
	def _escape(text):
		text = text.replace("-", "_")
		text = text.replace("/", "_")
		return text
	item_dict = { _escape(item.name): item for item in items }
	return LookupEnum(enum_name, item_dict)

Derating = collections.namedtuple("Derating", [ "security_lvl_bits", "reason" ])

HashFunction = collections.namedtuple("HashFunction", [ "name", "pretty_name", "output_bits", "derating", "oid" ])
HashFunctions = _to_enum("HashFunctions", [
	HashFunction(name = "md2", pretty_name = "MD2", output_bits = 128, derating = Derating(security_lvl_bits = 0, reason = "Søren S. Thomsen (2008). \"An improved preimage attack on MD2\""), oid = None),
	HashFunction(name = "md4", pretty_name = "MD4", output_bits = 128, derating = Derating(security_lvl_bits = 0, reason = "Yu Sasaki; et al. (2007). \"New message difference for MD4\""), oid = None),
	HashFunction(name = "md5", pretty_name = "MD5", output_bits = 128, derating = Derating(security_lvl_bits = 0, reason = "Xie Tao; Fanbao Liu & Dengguo Feng (2013). \"Fast Collision Attack on MD5\""), oid = None),
	HashFunction(name = "sha0", pretty_name = "SHA0", output_bits = 160, derating = Derating(security_lvl_bits = 0, reason = "Manuel, Stéphane; Peyrin, Thomas (2008). \"Collisions on SHA-0 in One Hour\""), oid = None),
	HashFunction(name = "sha1", pretty_name = "SHA-1", output_bits = 160, derating = Derating(security_lvl_bits = 0, reason = "Stevens, Marc; Bursztein, Elie; Karpman, Pierre; Albertini, Ange; Markov, Yarik. \"The first collision for full SHA-1\""), oid = OIDDB.HashFunctions.inverse("sha1")),
	HashFunction(name = "sha224", pretty_name = "SHA-224", output_bits = 224, derating = None, oid = OIDDB.HashFunctions.inverse("sha224")),
	HashFunction(name = "sha256", pretty_name = "SHA-256", output_bits = 256, derating = None, oid = OIDDB.HashFunctions.inverse("sha256")),
	HashFunction(name = "sha384", pretty_name = "SHA-384", output_bits = 384, derating = None, oid = OIDDB.HashFunctions.inverse("sha384")),
	HashFunction(name = "sha512", pretty_name = "SHA-512", output_bits = 512, derating = None, oid = OIDDB.HashFunctions.inverse("sha512")),
	HashFunction(name = "sha3-224", pretty_name = "SHA3-224", output_bits = 224, derating = None, oid = OIDDB.HashFunctions.inverse("sha3-224")),
	HashFunction(name = "sha3-256", pretty_name = "SHA3-256", output_bits = 256, derating = None, oid = OIDDB.HashFunctions.inverse("sha3-256")),
	HashFunction(name = "sha3-384", pretty_name = "SHA3-384", output_bits = 384, derating = None, oid = OIDDB.HashFunctions.inverse("sha3-384")),
	HashFunction(name = "sha3-512", pretty_name = "SHA3-512", output_bits = 512, derating = None, oid = OIDDB.HashFunctions.inverse("sha3-512")),
	HashFunction(name = "shake256-912", pretty_name = "Shake256-912", output_bits = 912, derating = None, oid = None),
])

Cryptosystem = collections.namedtuple("Cryptosystem", [ "name", "shortcut", "spec_parameters" ])
Cryptosystems = _to_enum("Cryptosystems", [
	Cryptosystem(name = "RSA", shortcut = "rsa", spec_parameters = [ ("bitlen", int) ]),
	Cryptosystem(name = "DSA", shortcut = "dsa", spec_parameters = [ ("L", int), ("N", int) ]),		# L is modulus bitlength, N is bitlenght of q
	Cryptosystem(name = "ECC/ECDSA", shortcut = "ecc", spec_parameters = [ ("curvename", str) ]),
	Cryptosystem(name = "ECC/EdDSA", shortcut = "eddsa", spec_parameters = [ ("curvename", str) ]),
	Cryptosystem(name = "ECC/ECDH", shortcut = "ed-ecdh", spec_parameters = [ ("curvename", str) ]),
])

SignatureFunction = collections.namedtuple("SignatureFunction", [ "name", "pretty_name", "cryptosystem" ])
SignatureFunctions = _to_enum("SignatureFunctions", [
	SignatureFunction(name = "rsa-encryption", pretty_name = "RSA encryption", cryptosystem = Cryptosystems.RSA),
	SignatureFunction(name = "rsa-ssa-pss", pretty_name = "RSA/PSS", cryptosystem = Cryptosystems.RSA),
	SignatureFunction(name = "dsa", pretty_name = "DSA", cryptosystem = Cryptosystems.DSA),
	SignatureFunction(name = "ecdsa", pretty_name = "ECDSA", cryptosystem = Cryptosystems.ECC_ECDSA),
	SignatureFunction(name = "eddsa", pretty_name = "EdDSA", cryptosystem = Cryptosystems.ECC_EdDSA),
])

EdDSAParams = collections.namedtuple("EdDSAParams", [ "curve", "prehash" ])

SignatureAlgorithm = collections.namedtuple("SignatureAlgorithm", [ "name", "hash_fnc", "sig_fnc", "sig_params", "oid" ])
SignatureAlgorithms = _to_enum("SignatureAlgorithms", [
	SignatureAlgorithm(name = "md2WithRsaEncryption", hash_fnc = HashFunctions.md2, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = [ OIDDB.SignatureAlgorithms.inverse("md2WithRsaEncryption"), OIDDB.SignatureAlgorithms.inverse("md2WithRsaEncryption-alt1") ] ),
	SignatureAlgorithm(name = "md4WithRsaEncryption", hash_fnc = HashFunctions.md4, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = [ OIDDB.SignatureAlgorithms.inverse("md4WithRsaEncryption"), OIDDB.SignatureAlgorithms.inverse("md4WithRsaEncryption-alt1") ]),
	SignatureAlgorithm(name = "md5WithRsaEncryption", hash_fnc = HashFunctions.md5, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = [ OIDDB.SignatureAlgorithms.inverse("md5WithRsaEncryption"), OIDDB.SignatureAlgorithms.inverse("md5WithRsaEncryption-alt1"), OIDDB.SignatureAlgorithms.inverse("md5WithRsaEncryption-alt2") ]),
	SignatureAlgorithm(name = "shaWithRsaEncryption", hash_fnc = HashFunctions.sha0, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("shaWithRsaEncryption")),
	SignatureAlgorithm(name = "sha1WithRsaEncryption", hash_fnc = HashFunctions.sha1, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = [ OIDDB.SignatureAlgorithms.inverse("sha1WithRsaEncryption"), OIDDB.SignatureAlgorithms.inverse("sha1WithRsaEncryption-alt1") ]),
	SignatureAlgorithm(name = "sha256WithRsaEncryption", hash_fnc = HashFunctions.sha256, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("sha256WithRsaEncryption")),
	SignatureAlgorithm(name = "sha384WithRsaEncryption", hash_fnc = HashFunctions.sha384, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("sha384WithRsaEncryption")),
	SignatureAlgorithm(name = "sha512WithRsaEncryption", hash_fnc = HashFunctions.sha512, sig_fnc = SignatureFunctions.rsa_encryption, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("sha512WithRsaEncryption")),
	SignatureAlgorithm(name = "dsa_with_SHA1", hash_fnc = HashFunctions.sha1, sig_fnc = SignatureFunctions.dsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("dsa-with-sha1")),
	SignatureAlgorithm(name = "dsa_with_SHA224", hash_fnc = HashFunctions.sha224, sig_fnc = SignatureFunctions.dsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("dsa-with-sha224")),
	SignatureAlgorithm(name = "dsa_with_SHA256", hash_fnc = HashFunctions.sha256, sig_fnc = SignatureFunctions.dsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("dsa-with-sha256")),
	SignatureAlgorithm(name = "dsa_with_SHA384", hash_fnc = HashFunctions.sha384, sig_fnc = SignatureFunctions.dsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("dsa-with-sha384")),
	SignatureAlgorithm(name = "dsa_with_SHA512", hash_fnc = HashFunctions.sha512, sig_fnc = SignatureFunctions.dsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("dsa-with-sha512")),
	SignatureAlgorithm(name = "ecdsa-with-SHA224", hash_fnc = HashFunctions.sha224, sig_fnc = SignatureFunctions.ecdsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("ecdsa-with-SHA224")),
	SignatureAlgorithm(name = "ecdsa-with-SHA256", hash_fnc = HashFunctions.sha256, sig_fnc = SignatureFunctions.ecdsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("ecdsa-with-SHA256")),
	SignatureAlgorithm(name = "ecdsa-with-SHA384", hash_fnc = HashFunctions.sha384, sig_fnc = SignatureFunctions.ecdsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("ecdsa-with-SHA384")),
	SignatureAlgorithm(name = "ecdsa-with-SHA512", hash_fnc = HashFunctions.sha512, sig_fnc = SignatureFunctions.ecdsa, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("ecdsa-with-SHA512")),
	SignatureAlgorithm(name = "RSASSA-PSS", hash_fnc = None, sig_fnc = SignatureFunctions.rsa_ssa_pss, sig_params = None, oid = OIDDB.SignatureAlgorithms.inverse("RSASSA-PSS")),
	SignatureAlgorithm(name = "Ed25519", hash_fnc = HashFunctions.sha512, sig_fnc = SignatureFunctions.eddsa, sig_params = EdDSAParams(curve = "Ed25519", prehash = False), oid = OIDDB.KeySpecificationAlgorithms.inverse("id-Ed25519")),
	SignatureAlgorithm(name = "Ed448", hash_fnc = HashFunctions.shake256_912, sig_fnc = SignatureFunctions.eddsa, sig_params = EdDSAParams(curve = "Ed448", prehash = False), oid = OIDDB.KeySpecificationAlgorithms.inverse("id-Ed448")),
])

PublicKeyAlgorithm = collections.namedtuple("PublicKeyAlgorithm", [ "name", "cryptosystem", "fixed_params", "oid" ])
PublicKeyAlgorithms = _to_enum("PublicKeyAlgorithms", [
	PublicKeyAlgorithm(name = "RSA", cryptosystem = Cryptosystems.RSA, fixed_params = None, oid = OIDDB.KeySpecificationAlgorithms.inverse("rsaEncryption")),
	PublicKeyAlgorithm(name = "DSA", cryptosystem = Cryptosystems.DSA, fixed_params = None, oid = OIDDB.KeySpecificationAlgorithms.inverse("id-dsa")),
	PublicKeyAlgorithm(name = "ECC", cryptosystem = Cryptosystems.ECC_ECDSA, fixed_params = None, oid = OIDDB.KeySpecificationAlgorithms.inverse("ecPublicKey")),
	PublicKeyAlgorithm(name = "Ed25519", cryptosystem = Cryptosystems.ECC_EdDSA, fixed_params = { "prehash": False }, oid = OIDDB.KeySpecificationAlgorithms.inverse("id-Ed25519")),
	PublicKeyAlgorithm(name = "Ed448", cryptosystem = Cryptosystems.ECC_EdDSA, fixed_params = { "prehash": False }, oid = OIDDB.KeySpecificationAlgorithms.inverse("id-Ed448")),
	PublicKeyAlgorithm(name = "X25519", cryptosystem = Cryptosystems.ECC_ECDH, fixed_params = { "prehash": True }, oid = OIDDB.KeySpecificationAlgorithms.inverse("id-X25519")),
	PublicKeyAlgorithm(name = "X448", cryptosystem = Cryptosystems.ECC_ECDH, fixed_params = { "prehash": True }, oid = OIDDB.KeySpecificationAlgorithms.inverse("id-X448")),
])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2019-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tls.Enums import TLSVersionRecordLayer, TLSVersionHandshake, ContentType, HandshakeType, CipherSuite, CompressionMethod, ExtensionType, ServerNameType, ECPointFormats, SupportedGroups, TLSAlertLevel, TLSAlertDescription
from x509sak.tls.Structure import Structure, VariableType, instantiate_member as IM

RecordLayerPkt = Structure(name = "RecordLayerPkt", members = [
	IM("content_type",			"uint8", enum_class = ContentType),
	IM("record_layer_version",	"uint16", enum_class = TLSVersionRecordLayer),
	IM("payload",				"opaque16"),
])

TLSExtension = Structure(name = "TLSExtension", members = [
	IM("extension_id",					"uint16", enum_class = ExtensionType),
	IM("content",						"opaque16"),
])

TLSExtensionFlag = Structure(name = "TLSExtensionFlag", members = [
	IM("extension_id",					"uint16", enum_class = ExtensionType),
	IM("content",						"opaque16", fixed_value = bytes()),
])

TLSExtensionServerNameIndication = Structure(name = "TLSExtensionServerNameIndication", members = [
	IM("extension_id",					"uint16", enum_class = ExtensionType, fixed_value = ExtensionType.server_name),
	IM("content",						"opaque16", inner = Structure([
		IM("server_name_list",				"opaque16", contains_array = True, inner = Structure([
			IM("server_name_type",				"uint8", enum_class = ServerNameType, fixed_value = ServerNameType.Hostname),
			IM("server_name",					"opaque16", string_encoding = "ascii"),
		])),
	])),
])
TLSExtensionServerNameIndication.create = lambda hostname: { "content": { "server_name_list": [ { "server_name": hostname } ] } }

TLSExtensionECPointFormats = Structure(name = "TLSExtensionECPointFormats", members = [
	IM("extension_id",					"uint16", enum_class = ExtensionType, fixed_value = ExtensionType.ec_point_formats),
	IM("content",						"opaque16", inner = Structure([
		IM("point_formats",					"opaque8", contains_array = True, inner =
			IM("point_format",					"uint8", enum_class = ECPointFormats),
		),
	])),
])

TLSExtensionSupportedGroups = Structure(name = "TLSExtensionSupportedGroups", members = [
	IM("extension_id",					"uint16", enum_class = ExtensionType, fixed_value = ExtensionType.supported_groups),
	IM("content",						"opaque16", inner = Structure([
		IM("groups",						"opaque16", contains_array = True, inner =
			IM("group",							"uint16", enum_class = SupportedGroups),
		),
	])),
])

AnyExtension = VariableType(name = "AnyExtension", possible_inner_classes = [
	TLSExtensionServerNameIndication,
	TLSExtensionECPointFormats,
	TLSExtensionSupportedGroups,
	TLSExtensionFlag,
	TLSExtension,
])

AlertPkt = Structure(name = "AlertPkt", members = [
	IM("alert_level",		"uint8", enum_class = TLSAlertLevel),
	IM("alert_description",	"uint8", enum_class = TLSAlertDescription),
])

ClientHelloPkt = Structure(name = "ClientHelloPkt", members = [
	IM("handshake_type",				"uint8", enum_class = HandshakeType, fixed_value = HandshakeType.ClientHello),
	IM("payload",						"opaque24", inner = Structure([
		IM("handshake_protocol_version",	"uint16", enum_class = TLSVersionHandshake),
		IM("client_random",					"array[32]"),
		IM("session_id",					"opaque8"),
		IM("cipher_suites",					"opaque16", contains_array = True, inner =
			IM("cipher_suite",					"uint16", enum_class = CipherSuite),
		),
		IM("compression_methods",			"opaque8", contains_array = True, inner =
			IM("compression_method",			"uint8", enum_class = CompressionMethod),
		),
		IM("extensions",					"opaque16", contains_array = True, inner = AnyExtension),
	])),
])

ServerHelloPkt = Structure(name = "ServerHelloPkt", members = [
	IM("handshake_type",				"uint8", enum_class = HandshakeType, fixed_value = HandshakeType.ServerHello),
	IM("payload",						"opaque24", inner = Structure([
		IM("handshake_protocol_version",	"uint16", enum_class = TLSVersionHandshake),
		IM("server_random",					"array[32]"),
		IM("session_id",					"opaque8"),
		IM("cipher_suite",					"uint16", enum_class = CipherSuite),
		IM("compression_method",			"uint8", enum_class = CompressionMethod),
		IM("extensions",					"opaque16", contains_array = True, inner = AnyExtension),
	])),
])

CertificatePkt = Structure(name = "Certificate", members = [
	IM("handshake_type",				"uint8", enum_class = HandshakeType, fixed_value = HandshakeType.Certificate),
	IM("payload",						"opaque24", inner = Structure([
		IM("certificates",					"opaque24", contains_array = True, inner =
			IM("certificate", "opaque24"),
		),
	])),
])

ClientHandshakeMessage = VariableType(name = "ClientHandshakeMessage", possible_inner_classes = [
	ClientHelloPkt,
	CertificatePkt,
])

ServerHandshakeMessage = VariableType(name = "ServerHandshakeMessage", possible_inner_classes = [
	ServerHelloPkt,
	CertificatePkt,
])



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2018 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import tempfile
from x509sak.tests import BaseTest, ResourceFileLoader

class CmdLineTestsGraphPool(BaseTest):
	def test_create_dot(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".dot") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as certfile:
			self._run_x509sak([ "graph", "--outfile", f.name, certfile ])
			with open(f.name) as f:
				dotfile = f.read()
			self.assertIn("digraph", dotfile)
			self.assertIn("bcade7ce", dotfile)

	def test_render_dot_png(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".png") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as certfile:
			self._run_x509sak([ "graph", "--outfile", f.name, certfile ])
			with open(f.name, "rb") as f:
				dotfile = f.read()
			self.assertTrue(dotfile.startswith(b"\x89PNG\r\n"))

	def test_render_dot_ps(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".ps") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as certfile:
			self._run_x509sak([ "graph", "--outfile", f.name, certfile ])
			with open(f.name, "rb") as f:
				dotfile = f.read()
			self.assertTrue(dotfile.startswith(b"%!PS-Adobe-3.0"))

	def test_render_dot_pdf(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".pdf") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as certfile:
			self._run_x509sak([ "graph", "--outfile", f.name, certfile ])
			with open(f.name, "rb") as f:
				dotfile = f.read()
			self.assertTrue(dotfile.startswith(b"%PDF"))

	def test_render_unknown_ext(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".xyz") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem") as certfile:
			self._run_x509sak([ "graph", "--outfile", f.name, certfile ], success_return_codes = [ 1 ])
			self._run_x509sak([ "graph", "--outfile", f.name, "--format", "pdf", certfile ])
			with open(f.name, "rb") as f:
				dotfile = f.read()
			self.assertTrue(dotfile.startswith(b"%PDF"))

	def test_render_multiple(self):
		with tempfile.NamedTemporaryFile(prefix = "graph_", suffix = ".dot") as f, ResourceFileLoader("certs/ok/johannes-bauer.com.pem", "certs/ok/johannes-bauer-intermediate.pem", "certs/ok/johannes-bauer-root.pem") as certfiles:
			self._run_x509sak([ "graph", "--outfile", f.name ] + list(certfiles))



import setuptools

with open("README.md") as f:
	long_description = f.read()

setuptools.setup(
	name = "x509sak",
	packages = setuptools.find_packages(),
	version = "0.0.4",
	license = "gpl-3.0",
	description = "X.509 Swiss Army Knife is a toolkit atop OpenSSL to ease generation of CAs and aid white-hat pentesting",
	long_description = long_description,
	long_description_content_type = "text/markdown",
	author = "Johannes Bauer",
	author_email = "joe@johannes-bauer.com",
	url = "https://github.com/johndoe31415/x509sak",
	download_url = "https://github.com/johndoe31415/x509sak/archive/v0.0.4.tar.gz",
	keywords = [ "x509", "certificate", "toolkit", "openssl", "pki", "pentesting", "pkcs11", "ca" ],
	install_requires = [
		"pyasn1",
		"pyasn1_modules",
	],
	entry_points = {
		"console_scripts": [
			"x509sak = x509sak.__main__:main"
		]
	},
	include_package_data = True,
	classifiers = [
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Developers",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3",
		"Programming Language :: Python :: 3 :: Only",
		"Programming Language :: Python :: 3.5",
		"Programming Language :: Python :: 3.6",
		"Programming Language :: Python :: 3.7",
		"Programming Language :: Python :: 3.8",
		"Programming Language :: Python :: 3.9",
		"Programming Language :: Python :: 3.10",
	],
)



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate.Validator import BaseValidationResult, BaseValidator
from x509sak.estimate.GeneralNameValidator import GeneralNameValidator

class NameConstraintsSubtreeValidationResult(BaseValidationResult):

	def _validate_subtree(self, subtree):
		if subtree.minimum != 0:
			self._report("X509Cert_Body_X509Exts_Ext_NC_SubtreeTemplate_MinimumNotZero", "contains minimum value that is not zero, but %d." % (subtree.minimum))
		if subtree.maximum is not None:
			self._report("X509Cert_Body_X509Exts_Ext_NC_SubtreeTemplate_MaximumPresent", "contains maximum value that is present and %d." % (subtree.maximum))

		self._delegate(self._validator.base_name_validator, subtree.base)

	def _validate(self):
		for subtree in self._subject:
			self._validate_subtree(subtree)

class NameConstraintsSubtreeValidator(BaseValidator):
	_ValidationResultClass = NameConstraintsSubtreeValidationResult

	def __init__(self, validation_subject, recognized_issues):
		BaseValidator.__init__(self, validation_subject, recognized_issues)
		self._base_name_validator = GeneralNameValidator(validation_subject = validation_subject, recognized_issues = recognized_issues, allow_dnsname_wildcard_matches = True, ip_addresses_are_subnets = True, permissible_uri_schemes = [ "http", "https", "ldap" ])

	@property
	def base_name_validator(self):
		return self._base_name_validator



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import pyasn1
from x509sak.ModulusDB import ModulusDB
from x509sak.NumberTheory import NumberTheory
from x509sak.estimate.BaseEstimator import BaseEstimator
from x509sak.estimate import JudgementCode, AnalysisOptions, Verdict, Commonness, Compatibility
from x509sak.estimate.Judgement import SecurityJudgement, SecurityJudgements, RFCReference
from x509sak.Exceptions import LazyDeveloperException

@BaseEstimator.register
class RSASecurityEstimator(BaseEstimator):
	_ALG_NAME = "rsa"

	def __init__(self, analysis_options = None):
		super().__init__(analysis_options = analysis_options)
		if self._analysis_options.rsa_testing == AnalysisOptions.RSATesting.Full:
			self._test_probable_prime = True
			self._pollards_rho_iterations = 10000
		elif self._analysis_options.rsa_testing == AnalysisOptions.RSATesting.Some:
			self._test_probable_prime = False
			self._pollards_rho_iterations = 5
		elif self._analysis_options.rsa_testing == AnalysisOptions.RSATesting.Fast:
			self._test_probable_prime = False
			self._pollards_rho_iterations = 0
		else:
			raise LazyDeveloperException(NotImplemented, self._analysis_options.rsa_testing)

	@staticmethod
	def analyze_e(e):
		if e < 0:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_Negative, "RSA exponent is negative, this is either a wrong encoding or might be a malicious key.", bits = 0)
		elif e == 0:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_Zero, "RSA exponent is zero, this is a malicious key.", bits = 0)
		elif e == 1:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_One, "RSA exponent is 1, this is a malicious key.", bits = 0)
		elif e in [ 3, 5, 7, 17, 257 ]:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_Small, "RSA exponent is small, but fairly common.", verdict = Verdict.MEDIUM, commonness = Commonness.FAIRLY_COMMON)
		elif e < 65537:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_SmallAndUncommon, "RSA exponent is small and an uncommon choice.", verdict = Verdict.MEDIUM, commonness = Commonness.UNUSUAL)
		elif e == 65537:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_MostCommonValue, "RSA exponent is the most common choice.", verdict = Verdict.BEST_IN_CLASS, commonness = Commonness.COMMON)
		else:
			return SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Exponent_Large, "RSA exponent is uncommonly large. This need not be a weakness, but is highly unusual and may cause interoperability issues.", verdict = Verdict.BEST_IN_CLASS, commonness = Commonness.HIGHLY_UNUSUAL)

	def analyze_n(self, n):
		judgements = SecurityJudgements()

		if n < 0:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_Negative, "Modulus uses incorrect encoding, representation is a negative integer.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

			# Fix up n so it's a positive integer for the rest of the tests
			bitlen = (n.bit_length() + 7) // 8 * 8
			mask = (1 << bitlen) - 1
			n = n & mask
		elif n == 0:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_Zero, "Modulus is zero, this is definitely a broken RSA public key.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)
		elif n == 1:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_One, "Modulus is one, this is definitely a broken RSA public key.", bits = 0, commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION)

		if self._test_probable_prime:
			if NumberTheory.is_probable_prime(n):
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_Prime, "Modulus is prime, not a compound integer as we would expect for RSA.", bits = 0)

		if self._pollards_rho_iterations > 0:
			small_factor = NumberTheory.pollard_rho(n, max_iterations = self._pollards_rho_iterations)
			if small_factor is not None:
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_Factorable, "Modulus has small factor (%d) and is therefore trivially factorable." % (small_factor), bits = 0)

		match = ModulusDB().find(n)
		if match is not None:
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_FactorizationKnown, "Modulus is known to be compromised: %s" % (match.text), bits = 0)

		hweight_analysis = NumberTheory.hamming_weight_analysis(n)
		if (hweight_analysis is not None) and (not hweight_analysis.plausibly_random):
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_Modulus_BitBiasPresent, "Modulus does not appear to be random. Expected a Hamming weight between %d and %d for a %d bit modulus, but found Hamming weight %d." % (hweight_analysis.rnd_min_hweight, hweight_analysis.rnd_max_hweight, hweight_analysis.bitlen, hweight_analysis.hweight), commonness = Commonness.HIGHLY_UNUSUAL)

		# We estimate the complexity of factoring the modulus by the asymptotic
		# complexity of the GNFS.
		bits_security = NumberTheory.asymtotic_complexity_gnfs_bits(n)
		judgements += self.algorithm("bits").analyze(JudgementCode.X509Cert_PublicKey_RSA_Modulus_LengthInBits, bits_security)

		return judgements

	def analyze(self, pubkey):
		judgements = SecurityJudgements()
		result = {
			"cryptosystem":	"rsa",
			"specific": {
				"n": {
					"bits":		pubkey.n.bit_length(),
					"security":	self.analyze_n(pubkey.n),
				},
				"e": {
					"security":	self.analyze_e(pubkey.e),
				},
			},
			"security": judgements,
		}

		if pubkey.params is None:
			standard = RFCReference(rfcno = 3279, sect = "2.2.1", verb = "MUST", text = "When any of these three OIDs appears within the ASN.1 type AlgorithmIdentifier, the parameters component of that type SHALL be the ASN.1 type NULL.")
			judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_ParameterFieldNotPresent, "RSA parameter field should be present and should be of Null type, but is not present at all.", commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)
		else:
			# There is a parameters field present, it must be NULL
			(asn1_params, tail) = pyasn1.codec.der.decoder.decode(bytes(pubkey.params))
			if not isinstance(asn1_params, pyasn1.type.univ.Null):
				standard = RFCReference(rfcno = 3279, sect = "2.2.1", verb = "MUST", text = "When any of these three OIDs appears within the ASN.1 type AlgorithmIdentifier, the parameters component of that type SHALL be the ASN.1 type NULL.")
				judgements += SecurityJudgement(JudgementCode.X509Cert_PublicKey_RSA_RSAEncryption_ParameterFieldNotNULL, "RSA parameter field should be present and should be of Null type, but has different ASN.1 type %s." % (type(asn1_params).__name__), commonness = Commonness.HIGHLY_UNUSUAL, compatibility = Compatibility.STANDARDS_DEVIATION, standard = standard)

		judgements += result["specific"]["n"]["security"]
		judgements += result["specific"]["e"]["security"]

		if self._analysis_options.include_raw_data:
			result["specific"]["n"]["value"] = pubkey.n
			result["specific"]["e"]["value"] = pubkey.e
		return result



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.estimate.JudgementStructure import create_judgement_code_class
JudgementCode = create_judgement_code_class(verbose = False)

from x509sak.estimate.AnalysisOptions import AnalysisOptions
from x509sak.estimate.BaseEstimator import BaseEstimator as SecurityEstimator
from x509sak.estimate.Judgement import SecurityJudgement, Verdict, Commonness, Compatibility, StandardDeviationType

import x509sak.estimate.EstimateBits
import x509sak.estimate.EstimateCertificateExtensions
import x509sak.estimate.EstimateCertificatePurpose
import x509sak.estimate.EstimateCertificate
import x509sak.estimate.EstimateCertificateValidity
import x509sak.estimate.EstimateDN
import x509sak.estimate.EstimateECC
import x509sak.estimate.EstimateHashFunction
import x509sak.estimate.EstimatePublicKey
import x509sak.estimate.EstimateRSA
import x509sak.estimate.EstimateDSA
import x509sak.estimate.EstimateSigFunction
import x509sak.estimate.EstimateSig
import x509sak.estimate.EstimateCARelation



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

from x509sak.tests import BaseTest
from x509sak.BijectiveDict import BijectiveDict

class BijectiveDictTests(BaseTest):
	def test_simple(self):
		bijd = BijectiveDict({
			"a":	"1",
			"b":	"2",
			"c":	"3",
		})
		self.assertEqual(bijd["a"], "1")
		self.assertEqual(bijd["b"], "2")
		self.assertEqual(bijd["c"], "3")
		with self.assertRaises(KeyError):
			_ = bijd["B"]

		self.assertEqual(bijd.inverse("3"), "c")
		self.assertEqual(bijd.inverse("2"), "b")
		self.assertEqual(bijd.inverse("1"), "a")
		with self.assertRaises(KeyError):
			bijd.inverse("4")

	def test_nonbijective(self):
		with self.assertRaises(Exception):
			BijectiveDict({
				"a":	"1",
				"b":	"2",
				"c":	"2",
			})

	def test_key_predicate(self):
		bijd = BijectiveDict({
			"a":	"1",
			"b":	"2",
			"c":	"3",
		}, key_predicate = lambda key: key.lower())
		self.assertEqual(bijd["b"], "2")
		self.assertEqual(bijd["B"], "2")

	def test_key_predicate_nonbijective(self):
		with self.assertRaises(Exception):
			BijectiveDict({
				"a":	"1",
				"A":	"2",
				"c":	"2",
			}, key_predicate = lambda key: key.lower())



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2020-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
from x509sak.Exceptions import InvalidInputException

class StringParseException(InvalidInputException): pass

class StringParser():
	class ParseCode(enum.IntEnum):
		RegularChar = 0
		ControlChar = 1

	def __init__(self, escape_chars, meta_character = "\\"):
		self._escape_chars = set(escape_chars)
		self._meta_char = meta_character
		self._escape_chars.add(meta_character)

	def parse(self, input_string):
		index = 0
		while index < len(input_string):
			next_char = input_string[index]
			if next_char == self._meta_char:
				index += 1
				if index == len(input_string):
					raise StringParseException("Got escape metacharacter at end of string, with nothing to follow.")
				escaped_char = input_string[index]
				if escaped_char not in self._escape_chars:
					raise StringParseException("Got unexpectedly an escaped character '%s'. Escaped are only %s." %	(escaped_char, ", ".join(self._escape_chars)))
				yield (self.ParseCode.RegularChar, escaped_char)
			elif next_char in self._escape_chars:
				yield (self.ParseCode.ControlChar, next_char)
			else:
				yield (self.ParseCode.RegularChar, next_char)
			index += 1

	def escape(self, parsed_string):
		result = [ ]
		for (char_type, char) in parsed_string:
			if (char_type == self.ParseCode.RegularChar) and (char in self._escape_chars):
				result.append("%s%s" % (self._meta_char, char))
			else:
				result.append(char)
		return "".join(result)

	def split(self, input_string, control_char, reassemble = False):
		split_data = [ ]
		split_char = (self.ParseCode.ControlChar, control_char)
		for next_char in self.parse(input_string):
			if next_char != split_char:
				# Just append
				if len(split_data) == 0:
					split_data.append([ ])
				split_data[-1].append(next_char)
			else:
				split_data.append([ ])
		if reassemble:
			split_data = [ self.escape(item) for item in split_data ]
		return split_data



#	x509sak - The X.509 Swiss Army Knife white-hat certificate toolkit
#	Copyright (C) 2018-2020 Johannes Bauer
#
#	This file is part of x509sak.
#
#	x509sak is free software; you can redistribute it and/or modify
#	it under the terms of the GNU General Public License as published by
#	the Free Software Foundation; this program is ONLY licensed under
#	version 3 of the License, later versions are explicitly excluded.
#
#	x509sak is distributed in the hope that it will be useful,
#	but WITHOUT ANY WARRANTY; without even the implied warranty of
#	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#	GNU General Public License for more details.
#
#	You should have received a copy of the GNU General Public License
#	along with x509sak; if not, write to the Free Software
#	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#	Johannes Bauer <JohannesBauer@gmx.de>

import enum
from x509sak.Tools import JSONTools
from x509sak.Exceptions import LazyDeveloperException
from x509sak.KwargsChecker import KwargsChecker
from x509sak.estimate import JudgementCode

class Verdict(enum.IntEnum):
	NO_SECURITY = 0
	BROKEN = 1
	WEAK = 2
	MEDIUM = 3
	HIGH = 4
	BEST_IN_CLASS = 5

class Commonness(enum.IntEnum):
	HIGHLY_UNUSUAL = 0
	UNUSUAL = 1
	FAIRLY_COMMON = 2
	COMMON = 3

class Compatibility(enum.IntEnum):
	STANDARDS_DEVIATION = 0
	LIMITED_SUPPORT = 1
	FULLY_COMPLIANT = 2

class StandardDeviationType(enum.IntEnum):
	RECOMMENDATION = 0
	VIOLATION = 1

class SecurityJudgement():
	def __init__(self, code, text, bits = None, verdict = None, commonness = None, compatibility = None, prefix_topic = False, standard = None, literature = None, info_payload = None):
		assert((code is None) or isinstance(code, JudgementCode))
		assert((bits is None) or isinstance(bits, (int, float)))
		assert((verdict is None) or isinstance(verdict, Verdict))
		assert((commonness is None) or isinstance(commonness, Commonness))
		assert((compatibility is None) or isinstance(compatibility, Compatibility))
		self._code = code
		self._text = text
		self._bits = bits
		self._verdict = verdict
		self._commonness = commonness
		self._compatibility = compatibility
		self._prefix_topic = prefix_topic
		self._standard = standard
		self._literature = literature
		self._info_payload = info_payload
		if self._bits == 0:
			if self._verdict is None:
				self._verdict = Verdict.NO_SECURITY
			if self._commonness is None:
				self._commonness = Commonness.HIGHLY_UNUSUAL

	@property
	def codeenum(self):
		return self._code

	@property
	def code(self):
		return self._code.value

	@property
	def text(self):
		return self._text

	@property
	def bits(self):
		return self._bits

	@property
	def verdict(self):
		return self._verdict

	@property
	def commonness(self):
		return self._commonness

	@property
	def compatibility(self):
		return self._compatibility

	@property
	def standard(self):
		return self._standard

	@property
	def literature(self):
		return self._literature

	@property
	def info_payload(self):
		return self._info_payload

	@classmethod
	def from_dict(cls, judgement_data):
		if "code" in judgement_data:
			code = getattr(JudgementCode, judgement_data["code"])
		else:
			code = None
		text = judgement_data["text"]
		bits = judgement_data.get("bits")
		verdict = judgement_data.get("verdict")
		if verdict is not None:
			verdict = Verdict(verdict["value"])
		commonness = judgement_data.get("commonness")
		if commonness is not None:
			commonness = Commonness(commonness["value"])
		compatibility = judgement_data.get("compatibility")
		if compatibility is not None:
			compatibility = Compatibility(compatibility["value"])
		standard = judgement_data.get("standard")
		if standard is not None:
			standard = StandardReference.from_dict(standard)
		literature = judgement_data.get("literature")
		if literature is not None:
			literature = LiteratureReference.from_dict(literature)
		return cls(code = code, text = text, bits = bits, verdict = verdict, commonness = commonness, compatibility = compatibility, standard = standard, literature = literature)

	def to_dict(self):
		result = {
			"code":				self.code.name,
			"topic":			self.code.topic,
			"short_text":		self.code.description,
			"text":				self.text,
			"bits":				self.bits,
			"verdict":			JSONTools.translate(self.verdict) if (self.verdict is not None) else None,
			"commonness":		JSONTools.translate(self.commonness) if (self.commonness is not None) else None,
			"compatibility":	JSONTools.translate(self.compatibility) if (self.compatibility is not None) else None,
			"standard":			self.standard.to_dict() if (self.standard is not None) else None,
			"literature":		self.literature.to_dict() if (self.literature is not None) else None,
		}
		return { key: value for (key, value) in result.items() if value is not None }

	def dump(self, indent = 0):
		indent_str = ("    " * indent)
		print("%s%s" % (indent_str, str(self)))

	def __str__(self):
		return "SecurityJudgement<%s / %s>" % (self.code, self.text)

class SecurityJudgements():
	def __init__(self):
		self._judgements = [ ]

	@staticmethod
	def _minof(items):
		result = None
		for item in items:
			if result is None:
				result = item
			elif item is not None:
				result = min(result, item)
		return result

	@property
	def uniform_topic(self):
		return len(set(security_judgement.code.topic for security_judgement in self)) in [ 0, 1 ]

	@property
	def bits(self):
		return self._minof(item.bits for item in self)

	@property
	def verdict(self):
		return self._minof(item.verdict for item in self)

	@property
	def commonness(self):
		return self._minof(item.commonness for item in self)

	@property
	def compatibility(self):
		return self._minof(item.compatibility for item in self)

	def summary_judgement(self):
		return SecurityJudgement(code = None, text = "Summary", bits = self.bits, verdict = self.verdict, commonness = self.commonness, compatibility = self.compatibility)

	def __iadd__(self, judgement):
		if judgement is None:
			# Simply ignore it.
			pass
		elif isinstance(judgement, SecurityJudgement):
			self._judgements.append(judgement)
		elif isinstance(judgement, SecurityJudgements):
			self._judgements += judgement
		else:
			raise NotImplementedError(judgement)
		return self

	@classmethod
	def from_dict(cls, judgements_data):
		judgements = cls()
		for judgement_data in judgements_data["components"]:
			judgements += SecurityJudgement.from_dict(judgement_data)
		return judgements

	def to_dict(self):
		result = {
			"bits":				self.bits,
			"verdict":			self.verdict,
			"commonness":		self.commonness,
			"compatibility":	self.compatibility,
			"components":		[ judgement.to_dict() for judgement in self._judgements ],
		}
		return { key: value for (key, value) in result.items() if value is not None }

	def dump(self, indent = 0):
		indent_str = ("    " * indent)
		print("%sSecurityJudgements" % (indent_str))
		for judgement in self._judgements:
			judgement.dump(indent + 1)

	def __len__(self):
		return len(self._judgements)

	def __iter__(self):
		return iter(self._judgements)

	def __getitem__(self, index):
		return self._judgements[index]

	def __str__(self):
		return "SecurityJudgements<%s>" % (", ".join(str(judgement) for judgement in self))

class StandardReference():
	_STD_TYPE = None
	_REGISTERED = { }

	@classmethod
	def from_dict(cls, data):
		if data["type"] not in cls._REGISTERED:
			raise LazyDeveloperException("Class not registered for standards type '%s'." % (data["type"]))
		return cls._REGISTERED[data["type"]].from_dict(data)

	@property
	def deviation_type(self):
		raise NotImplementedError(self.__class__.__name__)

	@classmethod
	def register(cls, decoree):
		assert(decoree._STD_TYPE is not None)
		cls._REGISTERED[decoree._STD_TYPE] = decoree
		return decoree

@StandardReference.register
class RFCReference(StandardReference):
	_STD_TYPE = "RFC"

	def __init__(self, rfcno, sect, verb, text):
		assert(verb in [ "SHOULD", "MUST", "RECOMMEND", "MAY", "SHALL" ])
		StandardReference.__init__(self)
		self._rfcno = rfcno
		self._sect = sect
		self._verb = verb
		self._text = text

	@property
	def deviation_type(self):
		return {
			"SHOULD":		StandardDeviationType.RECOMMENDATION,
			"RECOMMEND":	StandardDeviationType.RECOMMENDATION,
			"MAY":			StandardDeviationType.RECOMMENDATION,
			"MUST":			StandardDeviationType.VIOLATION,
			"SHALL":		StandardDeviationType.VIOLATION,
		}[self.verb]

	@property
	def rfcno(self):
		return self._rfcno

	@property
	def sect(self):
		return self._sect

	@property
	def verb(self):
		return self._verb

	@property
	def text(self):
		return self._text

	@classmethod
	def from_dict(cls, data):
		return cls(rfcno = data["rfcno"], sect = data["sect"], verb = data["verb"], text = data["text"])

	def to_dict(self):
		return {
			"type":				self._STD_TYPE,
			"rfcno":			self.rfcno,
			"sect":				self.sect,
			"verb":				self.verb,
			"text":				self.text,
			"deviation_type":	self.deviation_type,
		}

	def __str__(self):
		if isinstance(self.sect, str):
			return "RFC%d Sect. %s" % (self.rfcno, self.sect)
		else:
			return "RFC%d Sects. %s" % (self.rfcno, " / ".join(self.sect))

@StandardReference.register
class LiteratureReference(StandardReference):
	_STD_TYPE = "literature"
	_Arguments = KwargsChecker(required_arguments = set([ "author", "title" ]), optional_arguments = set([ "type", "year", "month", "source", "quote", "doi", "sect", "deviation_type" ]), check_functions = {
		"year":		lambda x: isinstance(x, int),
		"month":	lambda x: isinstance(x, int) and (1 <= x <= 12),
	})
	_MONTHNAMES = {
		1:	"Jan",
		2:	"Feb",
		3:	"Mar",
		4:	"Apr",
		5:	"May",
		6:	"Jun",
		7:	"Jul",
		8:	"Aug",
		9:	"Sep",
		10:	"Oct",
		11:	"Nov",
		12:	"Dec",
	}


	def __init__(self, **kwargs):
		StandardReference.__init__(self)
		self._Arguments.check(kwargs, "LiteratureReference")
		self._fields = kwargs
		if isinstance(self._fields["author"], str):
			self._fields["author"] = [ self._fields["author"] ]
		self._fields["type"] = self._STD_TYPE

	@property
	def deviation_type(self):
		if "deviation_type" not in self._fields:
			return None
		else:
			return {
				"SHOULD":		StandardDeviationType.RECOMMENDATION,
				"MUST":			StandardDeviationType.VIOLATION,
			}[self._fields["deviation_type"]]

	@classmethod
	def from_dict(cls, data):
		return cls(**data)

	def to_dict(self):
		return dict(self._fields)

	def __str__(self):
		text = " and ".join(self._fields["author"])
		if "year" in self._fields is not None:
			if "month" in self._fields is not None:
				text += " (%s %d)" % (self._MONTHNAMES[self._fields["month"]], self._fields["year"])
			else:
				text += " (%d)" % (self._fields["year"])
		text += ". \"%s\"" % (self._fields["title"])
		if "sect" in self._fields:
			text += " Sect. %s" % (self._fields["sect"])
		return text
